#! /usr/bin/env python3

"A basic test script for the NKT Koheras Boostik laser amplifier."

import argparse
import qmi

from qmi.utils.context_managers import start_stop, open_close
from qmi.instruments.nkt_photonics.boostik import KoherasBoostikLaserAmplifier

boostik_transports = {
        "node1" : "serial:/dev/serial/by-id/usb-FTDI_FT232R_USB_UART_A1070VQD-if00-port0",
        "node2" : "serial:/dev/serial/by-id/usb-FTDI_FT232R_USB_UART_A106T0T3-if00-port0"
    }

def _dump_boostik_status(boostik: KoherasBoostikLaserAmplifier):
    print("amplifier information .......... : {}".format(boostik.get_amplifier_information()))
    print("current setpoint ............... : {:10.2f} [A]".format(boostik.get_current_setpoint()))
    print("actual current ................. : {:10.2f} [A]".format(boostik.get_actual_current()))
    print("diode booster temperature ...... : {:10.2f} [°C]".format(boostik.get_diode_booster_temperature()))
    print("ambient temperature ............ : {:10.2f} [°C]".format(boostik.get_ambient_temperature()))
    print("input power .................... : {:10.2f} [?]".format(boostik.get_input_power()))
    print("amplifier enabled .............. : {} [bool]".format(boostik.get_amplifier_enabled()))

def main():

    parser = argparse.ArgumentParser(description="Basic test of the Koheras Boostik laser amplifier.")
    parser.add_argument("node", choices = ["node1", "node2"], help="use the Boostik connected to this node")

    args = parser.parse_args()

    with start_stop(qmi, "test_boostik_driver"):
        boostik_transport = boostik_transports[args.node]
        boostik = qmi.make_instrument("boostik", KoherasBoostikLaserAmplifier, boostik_transport)
        boostik.open()
        try:
            _dump_boostik_status(boostik)
        finally:
            boostik.close()

if __name__ == "__main__":
    main()



"""
Time Base, AOM driver.

The qmi.instruments.timebase package provides support for:
- DIM3000 AOM driver.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.timebase.dim3000 import TimeBase_DIM3000 as TimeBase_Dim3000



import ctypes
import time
import unittest
from unittest.mock import patch

from qmi.core.exceptions import QMI_InstrumentException, QMI_InvalidOperationException
from qmi.instruments.picoquant.support._hhlib_function_signatures import _hhlib_function_signatures
from qmi.instruments.picoquant import PicoQuant_HydraHarp400

from tests.patcher import PatcherQmiContext as QMI_Context


class HydraHarpOldLibrayTestCase(unittest.TestCase):

    def test_initialize_excepts_valid_mode_but_too_old_lib(self):
        patcher = patch('ctypes.cdll.LoadLibrary', spec=ctypes.CDLL)
        self._library_mock = patcher.start().return_value

        function_names, _, _ = zip(*_hhlib_function_signatures)
        self._library_mock.mock_add_spec(function_names, spec_set=True)

        self.addCleanup(patcher.stop)

        self._hydraharp: PicoQuant_HydraHarp400 = PicoQuant_HydraHarp400(
            QMI_Context("Test_hydraharp"), 'hydraharp', '1111111'
        )

        self._library_mock.HH_GetLibraryVersion.return_value = 0
        self._library_mock.HH_OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'2.222')
        string_buffer2 = ctypes.create_string_buffer(b'1111111')
        retvals = (string_buffer, string_buffer2)

        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', side_effect=retvals):
            self._hydraharp.open()

        self._library_mock.HH_Initialize.return_value = 0
        with self.assertRaises(ValueError):
            self._hydraharp.initialize('T3', 'INTERNAL')

        self._library_mock.HH_CloseDevice.return_value = 0
        self._hydraharp.close()


class HydraHarpMethodsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        patcher = patch('ctypes.cdll.LoadLibrary', spec=ctypes.CDLL)
        self._library_mock = patcher.start().return_value

        function_names, _, _ = zip(*_hhlib_function_signatures)
        self._library_mock.mock_add_spec(function_names, spec_set=True)

        self.addCleanup(patcher.stop)

        self._hydraharp: PicoQuant_HydraHarp400 = PicoQuant_HydraHarp400(
            QMI_Context("test_hydraharp"), 'hydraharp', '1111111'
        )

        self._library_mock.HH_GetLibraryVersion.return_value = 0
        self._library_mock.HH_OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'1111111')

        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', return_value=string_buffer):
            self._hydraharp.open()

    def tearDown(self) -> None:
        self._library_mock.HH_CloseDevice.return_value = 0
        self._hydraharp.close()

    def test_open_hh(self):
        """Test HydraHarp open where regular open fails and we need to get the SN with a query."""
        self._library_mock.HH_CloseDevice.return_value = 0
        self._hydraharp.close()

        self._library_mock.HH_GetLibraryVersion.return_value = 0
        self._library_mock.HH_OpenDevice.return_value = 0
        self._library_mock.HH_Initialize.return_value = 0
        self._library_mock.HH_GetSerialNumber.return_value = 0

        string_buffer_lib = ctypes.create_string_buffer(8)
        string_buffer_lib.value = b'1.23.4'
        string_buffer_sn1 = ctypes.create_string_buffer(8)
        string_buffer_sn1.value = b''
        string_buffer_sn2 = ctypes.create_string_buffer(8)
        string_buffer_sn2.value = b'1111111'

        with patch('sys.platform', 'linux1'):
            with patch('ctypes.create_string_buffer', side_effect=[string_buffer_lib, string_buffer_sn1, string_buffer_sn2]):
                self._hydraharp.open()

            self._library_mock.HH_Initialize.assert_called_once()
            self._library_mock.HH_GetSerialNumber.assert_called_once()

    def test_initialize(self):
        self._library_mock.HH_Initialize.return_value = 0
        self._hydraharp.initialize('T2', 'INTERNAL')

    def test_initialize_excepts_measurement_running(self):
        self._library_mock.HH_Initialize.return_value = 0
        self._library_mock.HH_StartMeas.return_value = 0
        self._hydraharp.start_measurement(1000)
        with self.assertRaises(QMI_InvalidOperationException):
            self._hydraharp.initialize('T2', 'INTERNAL')

    def test_initialize_excepts_invalid_mode(self):
        self._library_mock.HH_Initialize.return_value = 0
        with self.assertRaises(ValueError):
            self._hydraharp.initialize('T4', 'INTERNAL')

    def test_initialize_raises_libraryError(self):

        self._library_mock.HH_Initialize.return_value = -1
        self._library_mock.HH_GetErrorString.return_value = 0

        error_string = 'error string'
        ctype_err_string = ctypes.create_string_buffer(error_string.encode('ASCII'), 40)

        with self.assertRaises(QMI_InstrumentException), \
                patch('ctypes.create_string_buffer', return_value=ctype_err_string):

            self._hydraharp.initialize('T2', 'INTERNAL')

    def test_get_features(self):
        self._library_mock.HH_GetFeatures.return_value = 0
        bitpattern = 0x0001

        with patch('ctypes.c_int', return_value=ctypes.c_int(bitpattern)):
            features = self._hydraharp.get_features()

        self.assertIn('DLL', features)

    def test_module_info(self):
        self._library_mock.HH_GetNumOfModules.return_value = 0
        self._library_mock.HH_GetModuleInfo.return_value = 0
        modules = 2
        expected_module_info = [(1, 2), (3, 4)]
        values = [modules, 1, 2, 3, 4]
        patcher = patch('ctypes.c_int', side_effect=[ctypes.c_int(value) for value in values])

        with patcher:
            module_info = self._hydraharp.get_module_info()

        for actual, expected in zip(module_info, expected_module_info):
            self.assertTupleEqual(actual, expected)

    def test_get_debug_info(self):
        self._library_mock.HH_GetHardwareDebugInfo.return_value = 0
        expected_debug_info = 'foo bar baz'
        string_buffer = ctypes.create_string_buffer(expected_debug_info.encode('ASCII'), 65536)

        with patch('ctypes.create_string_buffer', return_value=string_buffer):
            debug_info = self._hydraharp.get_debug_info()

        self.assertEqual(debug_info, expected_debug_info)

    def test_set_measurement_control(self):
        self._library_mock.HH_SetMeasControl.return_value = 0
        self._hydraharp.set_measurement_control('SINGLESHOT_CTC', 'RISING', 'FALLING')
        self._library_mock.HH_SetMeasControl.assert_called_once_with(0, 0, 1, 0)

    def test_get_flags(self):
        self._library_mock.HH_GetFlags.return_value = 0
        flag_bitset = 0b01

        with patch('ctypes.c_int', return_value=ctypes.c_int(flag_bitset)):
            flags = self._hydraharp.get_flags()

        self.assertIn("OVERFLOW", flags)

    def test_get_warnings(self):
        self._library_mock.HH_GetWarnings.return_value = 0
        flag_bitset = 0b01

        with patch('ctypes.c_int', return_value=ctypes.c_int(flag_bitset)):
            flags = self._hydraharp.get_warnings()

        self.assertIn("SYNC_RATE_ZERO", flags)

    def test_set_sync_channel_offset(self):

        self._library_mock.HH_SetSyncChannelOffset.return_value = 0
        self._hydraharp.set_sync_channel_offset(99999)
        self._library_mock.HH_SetSyncChannelOffset.assert_called_once_with(0, 99999)

    def test_set_marker_enable(self):
        self._library_mock.HH_SetMarkerEnable.return_value = 0
        self._hydraharp.set_marker_enable(True, False, True, False)
        self._library_mock.HH_SetMarkerEnable.assert_called_once_with(0, 1, 0, 1, 0)

    def test_set_marker_holdoff_time(self):
        self._library_mock.HH_SetMarkerHoldoffTime.return_value = 0
        self._hydraharp.set_marker_holdoff_time(123456)
        self._library_mock.HH_SetMarkerHoldoffTime.assert_called_once_with(0, 123456)

    def test_calibrate(self):
        self._library_mock.HH_Calibrate.return_value = 0
        self._hydraharp.calibrate()
        self._library_mock.HH_Calibrate.assert_called_once_with(0)

    def test_set_sync_cfd(self):
        self._library_mock.HH_SetSyncCFD.return_value = 0
        self._hydraharp.set_sync_cfd(100, 30)
        self._library_mock.HH_SetSyncCFD.assert_called_once_with(0, 100, 30)

    def test_input_cfd(self):
        self._library_mock.HH_SetInputCFD.return_value = 0
        self._hydraharp.set_input_cfd(2, 100, 30)
        self._library_mock.HH_SetInputCFD.assert_called_once_with(0, 2, 100, 30)

    def test_read_fifo(self):
        self._library_mock.HH_ReadFiFo.return_value = 0
        self._library_mock.HH_StartMeas.return_value = 0
        self._library_mock.HH_StopMeas.return_value = 0
        self._library_mock.HH_Initialize.return_value = 0

        self._hydraharp.initialize('T2', 'INTERNAL')
        self._hydraharp.start_measurement(1)
        time.sleep(0.01)
        self._hydraharp.stop_measurement()

        self._library_mock.HH_ReadFiFo.assert_called()
        self.assertEqual(len(self._library_mock.HH_ReadFiFo.call_args_list[0][0]), 4)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import Mock, patch, call
import logging

import qmi
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.pololu import Pololu_Maestro
from qmi.core.exceptions import QMI_InstrumentException


class PololuMaestroOpenCloseTestCase(unittest.TestCase):

    def setUp(self) -> None:
        qmi.start("pololu_unit_test")
        transport = "serial:COM1"
        self.instr = qmi.make_instrument("Pololu", Pololu_Maestro, transport)

    def tearDown(self) -> None:
        qmi.stop()

    def test_open_close(self):
        """Test opening and closing the instrument"""
        with patch("serial.Serial") as ser:
            self.instr.open()
            self.assertTrue(self.instr.is_open())
            self.instr.close()
            self.assertFalse(self.instr.is_open())
            ser.assert_called_once_with(
                "COM1",
                baudrate=9600,  # The rest are defaults
                bytesize=8,
                parity='N',
                rtscts=False,
                stopbits=1.0,
                timeout=0.04
            )


class PololuMaestroMinMaxTargetsConfigTestCase(unittest.TestCase):

    TRANSPORT_STR = "serial:/dev/ttyS1"
    CHANNEL1_MIN = 9000
    CHANNEL1_MAX = 9300
    CHANNEL3_MIN = 8000
    CHANNEL3_MAX = 9100

    def setUp(self) -> None:
        # suppress logging
        logging.getLogger("qmi.instruments.pololu.maestro").setLevel(
            logging.CRITICAL)
        self._cmd_lead = chr(0xAA) + chr(0x0C)
        self._error_check_cmd = bytes(self._cmd_lead + chr(0x21), "latin-1")
        qmi.start("pololu_unit_test")
        patcher = patch(
            'qmi.instruments.pololu.maestro.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        self.instr: Pololu_Maestro = qmi.make_instrument(
            "Pololu", Pololu_Maestro, self.TRANSPORT_STR,
            channels_min_max_targets={1: (self.CHANNEL1_MIN, self.CHANNEL1_MAX),
                                      3: (self.CHANNEL3_MIN, self.CHANNEL3_MAX)})
        self.instr.open()

    def tearDown(self) -> None:
        self.instr.close()
        qmi.stop()

    def test_setting_channel_min_and_max_targets(self):
        """Test setting the min and max targets of channels."""
        # Arrange
        channel = 1

        # Act
        actual_min = self.instr.get_min_target(channel)
        actual_max = self.instr.get_max_target(channel)

        # Assert
        self.assertEqual(actual_min, self.CHANNEL1_MIN)
        self.assertEqual(actual_max, self.CHANNEL1_MAX)

        # Arrange
        channel = 3

        # Act
        actual_min = self.instr.get_min_target(channel)
        actual_max = self.instr.get_max_target(channel)

        # Assert
        self.assertEqual(actual_min, self.CHANNEL3_MIN)
        self.assertEqual(actual_max, self.CHANNEL3_MAX)


class PololuMaestroMinMaxSpeedsConfigTestCase(unittest.TestCase):

    TRANSPORT_STR = "serial:/dev/ttyS1"
    CHANNEL1_MIN = 88
    CHANNEL1_MAX = 99

    def setUp(self) -> None:
        # suppress logging
        logging.getLogger("qmi.instruments.pololu.maestro").setLevel(
            logging.CRITICAL)
        self._cmd_lead = chr(0xAA) + chr(0x0C)
        self._error_check_cmd = bytes(self._cmd_lead + chr(0x21), "latin-1")
        qmi.start("pololu_unit_test")
        patcher = patch(
            'qmi.instruments.pololu.maestro.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        self.instr: Pololu_Maestro = qmi.make_instrument(
            "Pololu", Pololu_Maestro, self.TRANSPORT_STR,
            channels_min_max_speeds={1: (self.CHANNEL1_MIN, self.CHANNEL1_MAX)})
        self.instr.open()

    def tearDown(self) -> None:
        self.instr.close()
        qmi.stop()

    def test_setting_channel_min_and_max_speed(self):
        """Test setting the min and max speeds of channels."""
        # Arrange
        channel = 1

        # Act and Assert
        with self.assertRaises(ValueError):
            self.instr.set_speed(channel, 9000)


class PololuMaestroMinMaxAccelerationsConfigTestCase(unittest.TestCase):

    TRANSPORT_STR = "serial:/dev/ttyS1"
    CHANNEL1_MIN = 88
    CHANNEL1_MAX = 99

    def setUp(self) -> None:
        # suppress logging
        logging.getLogger("qmi.instruments.pololu.maestro").setLevel(
            logging.CRITICAL)
        self._cmd_lead = chr(0xAA) + chr(0x0C)
        self._error_check_cmd = bytes(self._cmd_lead + chr(0x21), "latin-1")
        qmi.start("pololu_unit_test")
        patcher = patch(
            'qmi.instruments.pololu.maestro.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        self.instr: Pololu_Maestro = qmi.make_instrument(
            "Pololu", Pololu_Maestro, self.TRANSPORT_STR,
            channels_min_max_accelerations={1: (self.CHANNEL1_MIN, self.CHANNEL1_MAX)})
        self.instr.open()

    def tearDown(self) -> None:
        self.instr.close()
        qmi.stop()

    def test_setting_channel_min_and_max_acceleration(self):
        """Test setting the min and max accelerations of channels."""
        # Arrange
        channel = 1

        # Act and Assert
        with self.assertRaises(ValueError):
            self.instr.set_acceleration(channel, 9000)


class PololuMaestroCommandsTestCase(unittest.TestCase):

    TRANSPORT_STR = "serial:/dev/ttyS1"

    def setUp(self) -> None:
        # suppress logging
        logging.getLogger("qmi.instruments.pololu.maestro").setLevel(
            logging.CRITICAL)
        self._cmd_lead = chr(0xAA) + chr(0x0C)
        self._error_check_cmd = bytes(self._cmd_lead + chr(0x21), "latin-1")
        qmi.start("pololu_unit_test")
        patcher = patch(
            'qmi.instruments.pololu.maestro.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        self.instr: Pololu_Maestro = qmi.make_instrument(
            "Pololu", Pololu_Maestro, self.TRANSPORT_STR)
        self.instr.open()

    def tearDown(self) -> None:
        self.instr.close()
        qmi.stop()

    def test_get_idn(self):
        """Test getting the QMI instrument ID."""
        # Arrange
        expected_vendor = "Pololu"
        expected_model = "Maestro Micro Servo Controller"

        # Act
        idn = self.instr.get_idn()

        # Assert
        self.assertEqual(expected_vendor, idn.vendor)
        self.assertEqual(expected_model, idn.model)
        self.assertIsNone(idn.serial)
        self.assertIsNone(idn.version)

    def test_set_target_value(self):
        """Test setting the target value."""
        # Arrange
        target, channel = 5000, 1
        lsb = target & 0x7F  # 7 bits for least significant byte
        msb = (target >> 7) & 0x7F  # shift 7 and take next 7 bits for msb
        cmd = chr(0x04) + chr(channel) + chr(lsb) + chr(msb)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        self._transport_mock.read.return_value = b'\x00\x00'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        self.instr.set_target(channel, target)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_target_enforces_min_values(self):
        """Test setting the target value below the minimum allowed value.
        The set value will be ignored and the target will be set to the minimum."""
        # Arrange
        target, channel = -20, 1
        cmd = chr(0x04) + chr(channel) + chr(0x00) + chr(0x00)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        self._transport_mock.read.return_value = b'\x00\x00'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        self.instr.set_target(channel, target)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_target_enforces_max_values(self):
        """Test setting the target value above the maximum allowed value.
        The set value will be ignored and the target will be set to the maximum."""
        # Arrange
        target, channel = 10000000, 1
        lsb = Pololu_Maestro.DEFAULT_MAX_VALUE & 0x7F
        msb = (Pololu_Maestro.DEFAULT_MAX_VALUE >> 7) & 0x7F
        cmd = chr(0x04) + chr(channel) + chr(lsb) + chr(msb)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        self._transport_mock.read.return_value = b'\x00\x00'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        self.instr.set_target(channel, target)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_get_target_values(self):
        """Test get target values return expected values."""
        # Arrange
        expected_initial = [0] * Pololu_Maestro.DEFAULT_NUM_CHANNELS

        # Act
        initial = self.instr.get_targets()

        # Assert
        self.assertListEqual(expected_initial, initial)

    def test_get_position(self):
        """Test get position returns expected value."""
        # Arrange
        expected, channel = 5000, 0
        cmd = chr(0x10) + chr(channel)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        low_bit, high_bit = expected - (expected // 256 * 256), expected // 256
        self._transport_mock.read.side_effect = [chr(low_bit),
                                                 chr(high_bit), b'\x00\x00']
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        result = self.instr.get_position(channel)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)
        self.assertEqual(result, expected)

    def test_set_speed(self):
        """Test setting a new speed for a channel."""
        # Arrange
        speed, channel = 500, 1
        lsb = speed & 0x7F  # 7 bits for least significant byte
        msb = (speed >> 7) & 0x7F  # shift 7 and take next 7 bits for msb
        cmd = chr(0x07) + chr(channel) + chr(lsb) + chr(msb)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        self._transport_mock.read.return_value = b'\x00\x00'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        self.instr.set_speed(channel, speed)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_speed_raises_exception(self):
        """Test setting an invalid speed for a channel raises an exception."""
        # Arrange
        speed, channel = 10000000, 1

        # Act and assert
        with self.assertRaises(ValueError):
            self.instr.set_speed(channel, speed)

    def test_set_acceleration(self):
        """Test setting a new acceleration for a channel."""
        # Arrange
        acceleration, channel = 50, 1
        lsb = acceleration & 0x7F  # 7 bits for least significant byte
        # shift 7 and take next 7 bits for msb
        msb = (acceleration >> 7) & 0x7F
        cmd = chr(0x09) + chr(channel) + chr(lsb) + chr(msb)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        self._transport_mock.read.return_value = b'\x00\x00'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        self.instr.set_acceleration(channel, acceleration)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_acceleration_raises_exception(self):
        """Test setting an invalid acceleration for a channel raises an exception."""
        acceleration, channel = 500, 1
        with self.assertRaises(ValueError):
            self.instr.set_acceleration(channel, acceleration)

    def test_go_home(self):
        """Test homing."""
        # Arrange
        expected_command = bytes(self._cmd_lead + chr(0x22), "latin-1")
        self._transport_mock.read.return_value = b'\x00\x00'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        self.instr.go_home()

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_get_set_max_target(self):
        """Test setting new maximum value on a channel."""
        # Arrange
        expected = 6100
        channel = 2

        # Act
        self.instr.set_max_target(channel, expected)

        # Assert
        actual = self.instr.get_max_target(channel)
        self.assertEqual(expected, actual)

    def test_set_max_target_excepts(self):
        """Test that setting new maximum value outside valid range excepts."""
        invalid_maxes = [1100020394, -999]
        channel = 4

        for m in invalid_maxes:
            with self.assertRaises(ValueError):
                self.instr.set_max_target(channel, m)

    def test_get_set_min_target(self):
        """Test setting new minimum value on a channel."""
        # Arrange
        expected = 7
        channel = 2

        # Act
        self.instr.set_min_target(channel, expected)

        # Assert
        actual = self.instr.get_min_target(channel)
        self.assertEqual(expected, actual)

    def test_set_min_target_excepts(self):
        """Test that setting new minimum value outside valid range excepts."""
        invalid_maxes = [1100020394, -999]
        channel = 4

        for m in invalid_maxes:
            with self.assertRaises(ValueError):
                self.instr.set_min_target(channel, m)

    def test_command_error_check_excepts(self):
        """Make a command's error check to except"""
        # Arrange
        cmd = chr(0x22)
        expected_command = bytes(self._cmd_lead + cmd, "latin-1")
        self._transport_mock.read.return_value = b'\x00\xF0'
        expected_calls = [
            call(expected_command),
            call(self._error_check_cmd)
        ]

        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.instr.go_home()

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_wrong_channel_excepts(self):
        """Test a set command with wrong channel numbers and see that it excepts"""
        wrong_channels = (-1, Pololu_Maestro.DEFAULT_NUM_CHANNELS)
        for ch in wrong_channels:
            with self.assertRaises(ValueError):
                self.instr.set_target(ch, 0)


if __name__ == '__main__':
    unittest.main()



"""
Teledyne FLIR, Point Grey; machine vision camera.

The qmi.instruments.ptgrey package provides support for:
- Blackfly digital camera, with
  - Aravis open-source module.
"""
# The class names are not fully QMI style compliant, but a distinction needs to be made.
from qmi.instruments.ptgrey.blackfly_aravis import PtGrey_BlackFly_Aravis as Flir_Blackfly_Aravis



"""Unit-tests for the New AG UC8 QMI driver."""

import logging

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast
from dataclasses import dataclass

from qmi.core.transport import QMI_Transport
from qmi.instruments.newport.ag_uc8 import Newport_AG_UC8
from qmi.instruments.newport.ag_uc8 import AxisStatus
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.exceptions import QMI_UsageException


# Disable all logging
logging.disable(logging.CRITICAL)


@dataclass
class TestMeta:
    """Test meta data."""

    transport: MagicMock
    transport_str: MagicMock
    name: MagicMock
    super: MagicMock
    instr: Newport_AG_UC8


class TestNewport_AG_UC8(TestCase):
    """Testcase for the Newport_AG_UC8 class."""

    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_name = MagicMock()
        mock_transport_str = MagicMock()
        mock_super = MagicMock()

        with patch(
            "qmi.instruments.newport.ag_uc8.create_transport",
            return_value=mock_transport,
        ):
            instr = Newport_AG_UC8(MagicMock(), mock_name, mock_transport_str)

        self._patcher_super = patch("qmi.instruments.newport.ag_uc8.super", mock_super)
        self._patcher_super.start()

        self._meta = TestMeta(
            transport=mock_transport,
            instr=cast(Newport_AG_UC8, instr),
            transport_str=mock_transport_str,
            name=mock_name,
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_init(self):
        """Newport_AG_UC8.__init__(), happy flow"""
        self.assertEqual(self._meta.instr._transport, self._meta.transport)

    def test_open(self):
        """Newport_AG_UC8.open(), happy flow"""

        self._meta.instr._write = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.open()
        self._meta.transport.open.assert_called_once_with()
        self._meta.super().open.assert_called_once_with()
        self._meta.instr._write.assert_called_once_with("MR")
        self._meta.instr._check_error.assert_called_once_with("MR")

    def test_close(self):
        """Newport_AG_UC8.close(), happy flow"""

        self._meta.instr.close()
        self._meta.transport.close.assert_called_once_with()
        self._meta.super().close.assert_called_once_with()

    def test_write(self):
        """Newport_AG_UC8._write(), happy flow"""

        mock_cmd = "COMMAND"
        self._meta.instr._check_is_open = MagicMock()

        with patch("qmi.instruments.newport.ag_uc8.time.sleep", MagicMock()):
            self._meta.instr._write(mock_cmd)

        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.transport.write.assert_called_once_with(
            mock_cmd.encode("ascii") + b"\r\n"
        )

    def test_ask(self):
        """Newport_AG_UC8._ask(), happy flow"""

        mock_timeout = MagicMock()
        mock_cmd = "COMMAND"
        mock_resp = "RESPONSE\r\n".encode("ascii")

        self._meta.instr._check_is_open = MagicMock()
        self._meta.transport.read_until = MagicMock(return_value=mock_resp)

        rt_val = self._meta.instr._ask(mock_cmd, mock_timeout)

        self._meta.transport.write.assert_called_once_with(
            mock_cmd.encode("ascii") + b"\r\n"
        )
        self._meta.transport.read_until.assert_called_once_with(
            message_terminator=b"\n", timeout=mock_timeout
        )
        self.assertEqual(rt_val, mock_resp.rstrip(b"\r\n").decode("ascii"))
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_attribute(self):
        """Newport_AG_UC8._get_attribute(), happy flow"""

        mock_cmd = "DL?"
        mock_timeout = MagicMock()
        mock_attr = 100
        mock_resp = f"DL {mock_attr}"
        self._meta.instr._ask = MagicMock(return_value=mock_resp)

        rt_val = self._meta.instr._get_attribute(mock_cmd, mock_timeout)
        self._meta.instr._ask.assert_called_once_with(mock_cmd, timeout=mock_timeout)
        self.assertEqual(rt_val, mock_attr)

    def test_get_attribute_invalid_response(self):
        """Newport_AG_UC8.get_attribute(), invalid response handling"""

        mock_cmd = "DL?"
        mock_timeout = MagicMock()
        mock_resp = f"INVALID"
        self._meta.instr._ask = MagicMock(return_value=mock_resp)

        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._get_attribute(mock_cmd, mock_timeout)
        self._meta.instr._ask.assert_called_once_with(mock_cmd, timeout=mock_timeout)

    def test_get_attribute_value_error(self):
        """Newport_AG_UC8.get_attribute(), value error handling"""

        mock_cmd = "DL?"
        mock_timeout = MagicMock()
        mock_resp = "DL INVALID"
        self._meta.instr._ask = MagicMock(return_value=mock_resp)

        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._get_attribute(mock_cmd, mock_timeout)
        self._meta.instr._ask.assert_called_once_with(mock_cmd, timeout=mock_timeout)

    def test_get_last_error(self):
        """Newport_AG_UC8.get_last_error(), happy flow"""

        mock_resp = MagicMock()
        self._meta.instr._get_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_last_error()
        self.assertEqual(rt_val, mock_resp)
        self._meta.instr._get_attribute.assert_called_once_with("TE")

    def test_check_error(self):
        """Newport_AG_UC8._check_error(), happy flow"""
        mock_cmd = MagicMock()
        mock_resp = 0
        self._meta.instr.get_last_error = MagicMock(return_value=mock_resp)
        self._meta.instr._check_error(mock_cmd)
        self._meta.instr.get_last_error.assert_called_once_with()

    def test_check_error_is_error(self):
        """Newport_AG_UC8._check_error(), happy flow"""
        mock_cmd = MagicMock()
        mock_resp = 1
        self._meta.instr.get_last_error = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._check_error(mock_cmd)

    def _test_get_axis_attribute(self, axis):
        mock_attr = MagicMock()
        mock_axis = axis
        mock_timeout = MagicMock()
        mock_resp = MagicMock()
        self._meta.instr._get_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._get_axis_attribute(
            mock_attr, mock_axis, mock_timeout
        )
        self._meta.instr._get_attribute.assert_called_once_with(
            f"{mock_axis}{mock_attr}", mock_timeout
        )
        self.assertEqual(rt_val, mock_resp)

    def test_get_axis_attribute(self):
        """Newport_AG_UC8._get_axis_attribute(), happy flow"""
        self._test_get_axis_attribute(1)
        self._test_get_axis_attribute(2)

    def test_get_axis_attribute_invalid_input(self):
        """Newport_AG_UC8._get_axis_attribute(), invalid input handling"""
        with self.assertRaises(QMI_UsageException):
            self._test_get_axis_attribute(0)

    def _test_set_axis_attribute(self, axis):
        mock_attr = MagicMock()
        mock_axis = axis
        mock_value = MagicMock()
        self._meta.instr._write = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr._set_axis_attribute(mock_attr, mock_axis, mock_value)
        self._meta.instr._write.assert_called_once_with(
            f"{mock_axis}{mock_attr}{mock_value}"
        )
        self._meta.instr._check_error.assert_called_once_with(
            f"{mock_axis}{mock_attr}{mock_value}"
        )

    def test_set_axis_attribute(self):
        """Newport_AG_UC8._set_axis_attribute(), happy flow"""
        self._test_set_axis_attribute(1)
        self._test_set_axis_attribute(2)

    def test_set_axis_attribute_invalid_input(self):
        """Newport_AG_UC8._set_axis_attribute(), invalid input handling"""
        with self.assertRaises(QMI_UsageException):
            self._test_set_axis_attribute(0)

    def test_reset(self):
        """Newport_AG_UC8.reset(), happy flow"""
        self._meta.instr._write = MagicMock()
        self._meta.instr._check_error = MagicMock()

        with patch("qmi.instruments.newport.ag_uc8.time.sleep", MagicMock()):
            self._meta.instr.reset()

        self._meta.instr._write.assert_has_calls(
            [
                call("RS"),
                call("MR"),
            ]
        )
        self._meta.instr._check_error.assert_called_once_with("MR")

    def test_get_idn(self):
        """Newport_AG_UC8.get_idn(), happy flow"""
        mock_resp = "MODEL\r\n VERSION \n"
        self._meta.instr._ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_idn()

        self.assertEqual(rt_val.vendor, "Newport")
        self.assertEqual(rt_val.model, "MODEL")
        self.assertEqual(rt_val.serial, None)
        self.assertEqual(rt_val.version, "VERSION")

    def test_get_idn_invalid_response(self):
        """Newport_AG_UC8.get_idn(), invalid response handling"""

        mock_resp = ""
        self._meta.instr._ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.get_idn()

    def _test_select_channel(self, channel):
        mock_channel = channel

        self._meta.instr._write = MagicMock()
        self._meta.instr._check_error = MagicMock()

        with patch("qmi.instruments.newport.ag_uc8.time.sleep", MagicMock()):
            self._meta.instr.select_channel(mock_channel)

        self._meta.instr._write.assert_called_once_with(f"CC{mock_channel}")
        self._meta.instr._check_error.assert_called_once_with("CC")

    def test_select_channel(self):
        """Newport_AG_UC8.select_channel(), happy flow"""
        self._test_select_channel(1)
        self._test_select_channel(2)
        self._test_select_channel(3)
        self._test_select_channel(4)

    def test_select_channel_same_channel(self):
        """Newport_AG_UC8.select_channel(), same channel no write"""
        self._test_select_channel(1)
        with self.assertRaises(AssertionError):
            self._test_select_channel(1)

    def test_select_channel_invalid_input(self):
        """Newport_AG_UC8.select_channel(), invalid input handling"""
        with self.assertRaises(QMI_UsageException):
            self._test_select_channel(0)

    def _test_get_limit_status(self, resp, result):
        mock_channel = MagicMock()
        mock_resp = resp
        self._meta.instr.select_channel = MagicMock()
        self._meta.instr._get_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_limit_status(mock_channel)
        self._meta.instr.select_channel.assert_called_once_with(mock_channel)
        self._meta.instr._get_attribute.assert_called_once_with("PH")
        self.assertEqual(rt_val, result)

    def test_get_limit_status(self):
        """Newport_AG_UC8.get_limit_status(), happy flow"""
        self._test_get_limit_status(0, (False, False))
        self._test_get_limit_status(1, (True, False))
        self._test_get_limit_status(2, (False, True))
        self._test_get_limit_status(3, (True, True))

    def test_get_limit_status_invalid_response(self):
        """Newport_AG_UC8.get_limit_status(), invalid response handling"""
        with self.assertRaises(QMI_InstrumentException):
            self._test_get_limit_status(4, (False, False))

    def test_get_step_delay(self):
        """Newport_AG_UC8.get_step_delay(), happy flow"""
        mock_resp = MagicMock()
        mock_axis = MagicMock()
        self._meta.instr._get_axis_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_step_delay(mock_axis)
        self._meta.instr._get_axis_attribute.assert_called_once_with("DL?", mock_axis)
        self.assertEqual(rt_val, mock_resp)

    def test_set_step_delay(self):
        """Newport_AG_UC8.set_step_delay(), happy flow"""
        mock_axis = MagicMock()
        mock_value = MagicMock()
        self._meta.instr._set_axis_attribute = MagicMock()
        self._meta.instr.set_step_delay(mock_axis, mock_value)
        self._meta.instr._set_axis_attribute.assert_called_once_with(
            "DL", mock_axis, mock_value
        )

    def _test_get_step_amplitude(self, direction, attr):
        mock_axis = MagicMock()
        mock_direction = direction
        mock_resp = MagicMock()

        self._meta.instr._get_axis_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_step_amplitude(mock_axis, mock_direction)
        self.assertEqual(rt_val, mock_resp)
        self._meta.instr._get_axis_attribute.assert_called_once_with(attr, mock_axis)

    def test_get_step_amplitude(self):
        """Newport_AG_UC8.get_step_amplitude(), happy flow"""
        self._test_get_step_amplitude(0, "SU+?")
        self._test_get_step_amplitude(1, "SU-?")

    def test_get_step_amplitude_invalid_input(self):
        """Newport_AG_UC8.get_step_amplitude(), invalid_input"""
        with self.assertRaises(QMI_UsageException):
            self._test_get_step_amplitude(3, "")

    def _test_set_step_amplitude(self, direction, attr):
        mock_axis = MagicMock()
        mock_direction = direction
        mock_resp = MagicMock()
        mock_value = MagicMock()

        self._meta.instr._set_axis_attribute = MagicMock(return_value=mock_resp)
        self._meta.instr.set_step_amplitude(mock_axis, mock_direction, mock_value)
        self._meta.instr._set_axis_attribute.assert_called_once_with(
            attr, mock_axis, mock_value
        )

    def test_set_step_amplitude(self):
        """Newport_AG_UC8.set_step_amplitude(), happy flow"""
        self._test_set_step_amplitude(0, "SU+")
        self._test_set_step_amplitude(1, "SU-")

    def test_set_step_amplitude_invalid_input(self):
        """Newport_AG_UC8.set_step_amplitude(), invalid_input"""
        with self.assertRaises(QMI_UsageException):
            self._test_set_step_amplitude(3, "")

    def test_get_step_count(self):
        """Newport_AG_UC8.get_step_count(), happy flow"""
        mock_axis = MagicMock()
        mock_resp = MagicMock()
        self._meta.instr._get_axis_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_step_count(mock_axis)
        self._meta.instr._get_axis_attribute.assert_called_once_with("TP", mock_axis)
        self.assertEqual(rt_val, mock_resp)

    def _test_clear_step_count(self, axis):
        mock_axis = axis
        self._meta.instr._write = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.clear_step_count(mock_axis)
        self._meta.instr._write.assert_called_once_with(f"{mock_axis}ZP")
        self._meta.instr._check_error.assert_called_once_with(f"{mock_axis}ZP")

    def test_clear_step_count(self):
        """Newport_AG_UC8.clear_step_count(), happy flow"""
        self._test_clear_step_count(1)
        self._test_clear_step_count(2)

    def test_clear_step_count_invalid_input(self):
        """Newport_AG_UC8.clear_step_count(), invalid input handling"""
        with self.assertRaises(QMI_UsageException):
            self._test_clear_step_count(3)

    def _test_get_axis_status(self, resp, status):
        mock_axis = MagicMock()
        mock_resp = resp
        self._meta.instr._get_axis_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_axis_status(mock_axis)
        self._meta.instr._get_axis_attribute.assert_called_once_with("TS", mock_axis)
        self.assertEqual(rt_val, status)

    def test_get_axis_status(self):
        """Newport_AG_UC8.get_axis_status(), happy flow"""
        self._test_get_axis_status(0, AxisStatus.READY)
        self._test_get_axis_status(1, AxisStatus.STEPPING)
        self._test_get_axis_status(2, AxisStatus.JOGGING)
        self._test_get_axis_status(3, AxisStatus.MOVING_TO_LIMIT)

    def test_set_jog(self):
        """Newport_AG_UC8.set_jog(), happy flow"""
        mock_channel = MagicMock()
        mock_axis = MagicMock()
        mock_speed = MagicMock()
        self._meta.instr.select_channel = MagicMock()
        self._meta.instr._set_axis_attribute = MagicMock()
        self._meta.instr.jog(mock_channel, mock_axis, mock_speed)
        self._meta.instr.select_channel.assert_called_once_with(mock_channel)
        self._meta.instr._set_axis_attribute.assert_called_once_with(
            "JA", mock_axis, mock_speed
        )

    def test_move_limit(self):
        """Newport_AG_UC8.move_limit(), happy flow"""
        mock_channel = MagicMock()
        mock_axis = MagicMock()
        mock_speed = MagicMock()
        self._meta.instr.select_channel = MagicMock()
        self._meta.instr._set_axis_attribute = MagicMock()
        self._meta.instr.move_limit(mock_channel, mock_axis, mock_speed)
        self._meta.instr.select_channel.assert_called_once_with(mock_channel)
        self._meta.instr._set_axis_attribute.assert_called_once_with(
            "MV", mock_axis, mock_speed
        )

    def test_measure_position(self):
        """Newport_AG_UC8.measure_position(), happy flow"""
        mock_channel = MagicMock()
        mock_axis = MagicMock()
        mock_resp = MagicMock()
        self._meta.instr.select_channel = MagicMock()
        self._meta.instr._get_axis_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.measure_position(mock_channel, mock_axis)
        self._meta.instr.select_channel.assert_called_once_with(mock_channel)
        self._meta.instr._get_axis_attribute.assert_called_once_with(
            "MA", mock_axis, timeout=self._meta.instr.SLOW_RESPONSE_TIMEOUT
        )
        self.assertEqual(rt_val, mock_resp)

    def test_move_abs(self):
        """Newport_AG_UC8.move_abs(), happy flow"""
        mock_channel = MagicMock()
        mock_axis = MagicMock()
        mock_position = MagicMock()
        mock_resp = MagicMock()
        self._meta.instr.select_channel = MagicMock()
        self._meta.instr._get_axis_attribute = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.move_abs(mock_channel, mock_axis, mock_position)
        self._meta.instr.select_channel.assert_called_once_with(mock_channel)
        self._meta.instr._get_axis_attribute.assert_called_once_with(
            f"PA{mock_position}",
            mock_axis,
            timeout=self._meta.instr.SLOW_RESPONSE_TIMEOUT,
        )
        self.assertEqual(rt_val, mock_resp)

    def test_move_rel(self):
        """Newport_AG_UC8.move_rel(), happy flow"""
        mock_channel = MagicMock()
        mock_axis = MagicMock()
        mock_steps = 0
        self._meta.instr.select_channel = MagicMock()
        self._meta.instr._set_axis_attribute = MagicMock()
        self._meta.instr.move_rel(mock_channel, mock_axis, mock_steps)
        self._meta.instr.select_channel.assert_called_once_with(mock_channel)
        self._meta.instr._set_axis_attribute.assert_called_once_with(
            "PR", mock_axis, mock_steps
        )

    def test_move_rel_invalid_input(self):
        """Newport_AG_UC8.move_rel(), invalid input handling"""
        with self.assertRaises(QMI_UsageException):
            self._meta.instr.move_rel(MagicMock(), MagicMock(), 2**31 + 1)

    def _test_stop(self, axis):
        mock_axis = axis
        self._meta.instr._write = MagicMock()
        self._meta.instr.stop(mock_axis)
        self._meta.instr._write.assert_called_once_with(f"{mock_axis}ST")

    def test_stop(self):
        """Newport_AG_UC8.stop(), happy flow"""
        self._test_stop(1)
        self._test_stop(2)

    def test_stop_invalid_input(self):
        """Newport_AG_UC8.stop(), invalid input handling"""
        with self.assertRaises(QMI_UsageException):
            self._test_stop(3)



"""Implementation of the QMI_Thread class.
"""

import threading

from qmi.core.exceptions import QMI_WrongThreadException


def check_in_main_thread() -> None:
    """Check that this function is called in the main thread.

    Raises:
        ~qmi.core.exceptions.QMI_WrongThreadException: If the function is called from any
            thread other than the main thread.
    """
    if threading.current_thread() is not threading.main_thread():
        raise QMI_WrongThreadException("Not in main thread")


class QMI_Thread(threading.Thread):
    """A QMI_Thread is an abstract base class for all threads used within QMI.

    The added value compared to Python's `threading.Thread` class is that it provides a
    way to request thread shutdown (via the `shutdown` method).

    QMI_Thread instances are initialized to be Python 'daemon' threads. This is done to prevent
    a deadlock when the Python process terminates; the Python interpreter effectively performs a
    'join()' on non-daemon threads at shutdown, prior even to running atexit handlers.

    Unfortunately, the shutdown process cannot be reliably hooked. By making the QMI_Threads
    daemonic, we ensure that we get to the execution of the at-exit handlers, which will properly
    tear down any remaining QMI_Contexts; This will shut down all active QMI_Threads in an orderly way.

    The QMI_Thread class currently has three specializations:

    - _EventDrivenThread in qmi.core.messaging;
    - _RpcThread in qmi.core.rpc;
    - _TaskThread in qmi.core.task.

    Important:
        QMI_Threads are part of the core QMI machinery.
        They should not be instantiated directly by QMI users.
        QMI users should use QMI_Tasks instead.
    """

    def __init__(self) -> None:
        super().__init__(daemon=True)
        self._shutdown_requested = False  # Indicates if thread shutdown has been requested.

    def run(self) -> None:
        """The run() method must be overridden by derived classes."""
        raise NotImplementedError()

    def shutdown(self) -> None:
        """Command the thread to terminate orderly, as soon as possible.

        After calling shutdown(), the usual thing to do is call join() to join up with the thread.

        The shutdown() function can also be called during Python cleanup, see the documentation of the join()
        method above.
        """

        # We check '_shutdown_requested', thereby guaranteeing that the _request_shutdown() method won't be called
        # more than once.
        if not self._shutdown_requested:
            self._shutdown_requested = True
            self._request_shutdown()

    def _request_shutdown(self) -> None:
        """Command the thread to terminate cleanly, as soon as possible.

        The _request_shutdown() method may be overridden by derived classes, so
        they can optionally take action to terminate their run() loop beyond polling
        the '_shutdown_requested' member variable.

        The way to initiate termination of the run() loop can differ between sub-classes.
        For example, it may involve waking up a running select() or wait().

        Simple implementations can just poll the '_shutdown_requested' member variable,
        which is fine for many use cases.

        The _request_shutdown() implementation _must_ be thread-safe, i.e., it
        should be possible to call the function from any thread.

        For a given QMI_Thread, the _request_shutdown() method will never be called
        more than once. This is guarded by the '_shutdown_requested' member variable.

        This function must only be called from the shutdown() method!
        """
        pass



#! /usr/bin/env python3

import logging
import math
import time
from typing import NamedTuple
import unittest
from unittest.mock import Mock, MagicMock

from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.context import QMI_Context
from qmi.core.exceptions import (
    QMI_MessageDeliveryException, QMI_UsageException, QMI_InvalidOperationException, QMI_DuplicateNameException
)
from qmi.core.rpc import (
    QMI_RpcObject, rpc_method, QMI_RpcTimeoutException, QMI_RpcFuture, QMI_RpcProxy, QMI_RpcNonBlockingProxy
)
from threading import Timer


class MyRpcTestClass(QMI_RpcObject):
    """An RPC test class"""
    _rpc_constants = ["CONSTANT_NUMBER"]

    CONSTANT_NUMBER = 42
    CONSTANT_FLOAT = 3.1415

    def __init__(self, context, name):
        super().__init__(context, name)

    @rpc_method
    def remote_sqrt(self, x):
        """This is some_method."""

        if x < 0:
            raise ValueError("Bad value!")

        # Sleep. This will be used to test timeout behaviour.
        time.sleep(x / 1e3)

        return math.sqrt(x)

    @rpc_method
    def my_lock_method(self, lock_token):
        """Method with it's own lock token keyword argument."""
        pass


class MyRpcSubClass(MyRpcTestClass):
    """An RPC sub class"""
    _rpc_constants = ["CONSTANT_STRING"]

    CONSTANT_STRING = "testing"

    @rpc_method
    def remote_log(self, x):
        return math.log(x)


class ProxyInterface(NamedTuple):
    """For testing proxy descriptor."""
    rpc_class_module: str = "SomeClass"
    rpc_class_name: str = "ClassyName"
    rpc_class_docstring: str = """This is Some Classy docstring."""
    constants: list = []
    methods: list = []
    signals: list = []


class ProxyDescriptor(NamedTuple):
    """For testing proxy descriptor."""
    address: MagicMock = MagicMock()
    category: MagicMock = MagicMock()
    interface: ProxyInterface = ProxyInterface()


class TestRpcProxy(unittest.TestCase):
    """Test instantiating QMI_RpcProxy and QMI_RpcNonBlockingProxy classes."""

    def test_create_instance(self):
        """Test creating an instance of QMI_RpcProxy."""
        # Arrange
        proxy_interface = ProxyInterface()
        expected_class_fqn = ".".join([proxy_interface.rpc_class_module, proxy_interface.rpc_class_name])
        expected_docstring = proxy_interface.rpc_class_docstring
        # Act
        proxy = QMI_RpcProxy(QMI_Context("test_rpcproxy"), ProxyDescriptor())
        # Assert
        self.assertIsInstance(proxy, QMI_RpcProxy)
        self.assertEqual(expected_class_fqn, proxy._rpc_class_fqn)
        self.assertEqual(expected_docstring, proxy.__doc__)

    def test_context_manager_excepts(self):
        """Test creating an instance of QMI_RpcProxy with context manager excepts with RecursionError."""
        # Act
        with self.assertRaises(RecursionError):
            with QMI_RpcProxy(QMI_Context("test_rpcproxy"), ProxyDescriptor()):
                pass

    def test_create_nonblocking_instance(self):
        """Test creating an instance of QMI_RpcNonBlockingProxy."""
        # Arrange
        proxy_interface = ProxyInterface()
        expected_class_fqn = ".".join([proxy_interface.rpc_class_module, proxy_interface.rpc_class_name])
        expected_docstring = proxy_interface.rpc_class_docstring
        # Act
        proxy = QMI_RpcNonBlockingProxy(QMI_Context("test_rpcnonblockingproxy"), ProxyDescriptor())
        # Assert
        self.assertIsInstance(proxy, QMI_RpcNonBlockingProxy)
        self.assertEqual(expected_class_fqn, proxy._rpc_class_fqn)
        self.assertEqual(expected_docstring, proxy.__doc__)

    def test_nonblocking_context_manager_excepts(self):
        """Test creating an instance of QMI_RpcProxy with context manager excepts with TypeError."""
        # Act
        with self.assertRaises((AttributeError, TypeError)):  # TypeError as no __enter__ and __exit__ present
            with QMI_RpcNonBlockingProxy(QMI_Context("test_rpcproxy"), ProxyDescriptor()):
                pass


class TestRPC(unittest.TestCase):

    def setUp(self):

        # Suppress warnings.
        logging.getLogger("qmi.core.rpc").setLevel(logging.CRITICAL)
        logging.getLogger("qmi.core.messaging").setLevel(logging.ERROR)

        # Start two contexts.
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0),
                "c2": CfgContext(tcp_server_port=0)
            }
        )

        c1 = QMI_Context("c1", config)
        c1.start()
        c1_port = c1.get_tcp_server_port()

        c2 = QMI_Context("c2", config)
        c2.start()

        # Connect c2 to c1.
        c1_address = "localhost:{}".format(c1_port)
        c2.connect_to_peer("c1", c1_address)

        self.c1 = c1
        self.c2 = c2

    def tearDown(self):

        self.c1.stop()
        self.c2.stop()

        self.c1 = None
        self.c2 = None

        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)
        logging.getLogger("qmi.core.messaging").setLevel(logging.NOTSET)

    def test_blocking_rpc(self):
        """Test for blocking RPC calls."""
        # Instantiate the class, as a thing to be serviced from context c1.
        # This gives us a proxy to the instance.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcTestClass)
        # Check nominal behavior.
        result = proxy1.remote_sqrt(256.0)
        self.assertEqual(result, 16.0)

        # Check exception behavior.
        with self.assertRaises(ValueError):
            proxy1.remote_sqrt(-1.0)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Check nominal behavior.
        result = proxy2.remote_sqrt(256.0)
        self.assertEqual(result, 16.0)
        # Assert also that docstring gets passed to proxy.
        self.assertEqual(MyRpcTestClass.__doc__, proxy1.__doc__)
        self.assertEqual(proxy1.__doc__, proxy2.__doc__)

        # Check exception behavior.
        with self.assertRaises(ValueError):
            proxy2.remote_sqrt(-1.0)

    def test_blocking_rpc_timeout(self):

        # Instantiate class in context c1.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcTestClass)

        # Check local timeout behavior.
        with self.assertRaises(QMI_RpcTimeoutException):
            proxy1.remote_sqrt(1024, rpc_timeout=1.0)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Check remote timeout behavior.
        with self.assertRaises(QMI_RpcTimeoutException):
            proxy2.remote_sqrt(1024, rpc_timeout=1.0)

    def test_nonblocking_rpc(self):

        # Instantiate the class, as a thing to be serviced from context c1.
        # This gives us a proxy to the instance.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcTestClass)

        # Check nominal behavior.
        future = proxy1.rpc_nonblocking.remote_sqrt(256.0)
        assert isinstance(future, QMI_RpcFuture)
        result = future.wait()
        self.assertEqual(result, 16.0)

        # Check exception behavior.
        with self.assertRaises(ValueError):
            future = proxy1.rpc_nonblocking.remote_sqrt(-1.0)
            assert isinstance(future, QMI_RpcFuture)
            future.wait()

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Check nominal behavior.
        future = proxy2.rpc_nonblocking.remote_sqrt(256.0)
        assert isinstance(future, QMI_RpcFuture)
        result = future.wait()
        self.assertEqual(result, 16.0)
        # Assert also that docstring gets passed to proxy.
        self.assertEqual(MyRpcTestClass.__doc__, proxy1.__doc__)
        self.assertEqual(proxy1.__doc__, proxy2.__doc__)

        # Check exception behavior.
        with self.assertRaises(ValueError):
            future = proxy2.rpc_nonblocking.remote_sqrt(-1.0)
            assert isinstance(future, QMI_RpcFuture)
            future.wait()

    def test_nonblocking_rpc_queue(self):

        # Instantiate class in context c1.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcTestClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Send several slow requests.
        future1 = proxy1.rpc_nonblocking.remote_sqrt(128)
        future2 = proxy2.rpc_nonblocking.remote_sqrt(129)
        future3 = proxy1.rpc_nonblocking.remote_sqrt(130)
        future4 = proxy1.rpc_nonblocking.remote_sqrt(131)
        future5 = proxy1.rpc_nonblocking.remote_sqrt(132)
        future6 = proxy1.rpc_nonblocking.remote_sqrt(133)

        # Wait for answers.
        result = future1.wait()
        self.assertAlmostEqual(result, math.sqrt(128))
        result = future2.wait()
        self.assertAlmostEqual(result, math.sqrt(129))

        # Wait out-of-order.
        result = future5.wait()
        self.assertAlmostEqual(result, math.sqrt(132))
        result = future3.wait(timeout=0.001)
        self.assertAlmostEqual(result, math.sqrt(130))
        result = future4.wait(timeout=0.001)
        self.assertAlmostEqual(result, math.sqrt(131))
        result = future6.wait()
        self.assertAlmostEqual(result, math.sqrt(133))

    def test_nonblocking_rpc_timeout(self):

        # Instantiate class in context c1.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcTestClass)

        # Check timeout behavior.
        with self.assertRaises(QMI_RpcTimeoutException):
            future = proxy1.rpc_nonblocking.remote_sqrt(1024)
            assert isinstance(future, QMI_RpcFuture)
            _ = future.wait(timeout=1.0)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Check timeout behavior.
        with self.assertRaises(QMI_RpcTimeoutException):
            future = proxy2.rpc_nonblocking.remote_sqrt(1024)
            assert isinstance(future, QMI_RpcFuture)
            _ = future.wait(timeout=1.0)

    def test_force_unlock(self):
        """Test locking the object in one proxy and force unlocking from another proxy."""
        # Instantiate class in context c1.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcTestClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        proxy1.lock()
        # See that second proxy cannot unlock
        result = proxy2.unlock()
        self.assertFalse(result)
        self.assertTrue(proxy1.is_locked())

        # Then, using force_unlock, we do succeed
        proxy2.force_unlock()
        self.assertFalse(proxy1.is_locked())

    def test_subclass(self):

        # Make instance of MyRpcSubClass in the first context.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Make an RPC call to a method defined by the subclass.
        result = proxy2.remote_log(10.0)
        self.assertAlmostEqual(result, 2.302585092994046)

        # Make an RPC call to a method defined by the base class.
        result = proxy2.remote_sqrt(100.0)
        self.assertAlmostEqual(result, 10.0)
        # Assert also that docstring gets passed to proxy.
        self.assertEqual(MyRpcSubClass.__doc__, proxy1.__doc__)
        self.assertEqual(proxy1.__doc__, proxy2.__doc__)

    def test_constants(self):

        # Make instance of MyRpcSubClass in the first context.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Check that constants are accessible via both proxies.
        self.assertEqual(proxy1.CONSTANT_NUMBER, 42)
        self.assertEqual(proxy1.CONSTANT_STRING, "testing")
        self.assertEqual(proxy2.CONSTANT_NUMBER, 42)
        self.assertEqual(proxy2.CONSTANT_STRING, "testing")

        # Check that non-exported constants are not accessible.
        with self.assertRaises(AttributeError):
            print(proxy1.CONSTANT_FLOAT)

    def test_call_to_disconnected(self):

        # Make instance of MyRpcSubClass in the first context.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Force the contexts to disconnect.
        self.c2.disconnect_from_peer("c1")

        # Check that RPC call within context c1 still works.
        result = proxy1.remote_sqrt(64.0)
        self.assertAlmostEqual(result, 8.0)

        # Check that RPC call between contexts fails.
        with self.assertRaises(QMI_MessageDeliveryException):
            proxy2.remote_sqrt(32.0)

    def test_call_to_removed_object(self):

        # Make instance of MyRpcSubClass in the first context.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Remove the object.
        self.c1.remove_rpc_object(proxy1)

        # Check that RPC call to removed object fails within context c1.
        with self.assertRaises(QMI_MessageDeliveryException):
            proxy1.remote_sqrt(16.0)

        # Check that RPC call to removed object fails even from context c2.
        with self.assertRaises(QMI_MessageDeliveryException):
            proxy2.remote_sqrt(32.0)

        # Check that non-blocking calls deliver the exception when waited on.
        future1 = proxy1.rpc_nonblocking.remote_sqrt(16.0)
        future2 = proxy2.rpc_nonblocking.remote_sqrt(32.0)
        with self.assertRaises(QMI_MessageDeliveryException):
            future1.wait()
        with self.assertRaises(QMI_MessageDeliveryException):
            future2.wait()

    def test_disconnect_during_call(self):

        # Make instance of MyRpcSubClass in the first context.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Start a slow, remote RPC call from the second context.
        future = proxy2.rpc_nonblocking.remote_sqrt(512)

        # While the call is in progress, force the contexts to disconnect.
        time.sleep(0.05)
        self.c2.disconnect_from_peer("c1")

        # Check that the RPC call completes with an error.
        with self.assertRaises(QMI_MessageDeliveryException):
            future.wait()

    def test_remove_object_while_call_queued(self):

        # Make instance of MyRpcSubClass in the first context.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)

        # Make a proxy via the second context.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")

        # Start a slow RPC call.
        future1 = proxy1.rpc_nonblocking.remote_sqrt(512)

        # Start two more RPC calls, one from each context.
        # These will be queued while the first call is still in progress.
        future2 = proxy1.rpc_nonblocking.remote_sqrt(15)
        future3 = proxy2.rpc_nonblocking.remote_sqrt(17)

        # Remove the RPC object while the first call is still in progress.
        time.sleep(0.05)
        self.c1.remove_rpc_object(proxy1)

        # Check that the first call completed successfully.
        result = future1.wait()
        self.assertAlmostEqual(result, math.sqrt(512))

        # Check that the other two calls complete with an error.
        with self.assertRaises(QMI_MessageDeliveryException):
            future2.wait()
        with self.assertRaises(QMI_MessageDeliveryException):
            future3.wait()

    def test_get_name_of_RpcObject(self):
        """Test `get_name()` rpc method of the class."""
        expected = "tc1"
        # Make instance of MyRpcSubClass in the first context and get its name.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)
        name1 = proxy1.get_name()
        # Make a proxy via the second context and get its name.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")
        name2 = proxy2.get_name()
        # Assert the names are as expected
        self.assertEqual(expected, name1)
        self.assertEqual(expected, name2)

    def test_get_address_of_RpcObject(self):
        """Test `get_address()` rpc method of the class."""
        expected = "c1.tc1"
        # Make instance of MyRpcSubClass in the first context and get its name.
        proxy1 = self.c1.make_rpc_object("tc1", MyRpcSubClass)
        # address1 = proxy1.get_address()
        address1 = proxy1.address
        # Make a proxy via the second context and get its address.
        proxy2 = self.c2.get_rpc_object_by_name("c1.tc1")
        # address2 = proxy2.get_address()
        address2 = proxy2.address
        # Assert the addresses are as expected
        self.assertEqual(expected, address1)
        self.assertEqual(expected, address2)

    def test_no_context_manager_allowed(self):
        """Making an RPC object with context manager is not allowed."""
        with self.assertRaises(NotImplementedError):
            with self.c1.make_rpc_object("tc1", MyRpcSubClass):
                pass


class TestRpcMethodDecorator(unittest.TestCase):
    class ObjectWithGoodMethodName(QMI_RpcObject):
        @rpc_method
        def some_method(self): ...

    class ObjectWithBadMethodNameLock(QMI_RpcObject):
        @rpc_method
        def lock(self): ...

    class ObjectWithBadMethodNameUnlock(QMI_RpcObject):
        @rpc_method
        def unlock(self): ...

    class ObjectWithBadMethodNameForceUnlock(QMI_RpcObject):
        @rpc_method
        def force_unlock(self): ...

    class ObjectWithBadMethodNameIsLocked(QMI_RpcObject):
        @rpc_method
        def is_locked(self): ...

    def test_good_name(self):
        """Test acceptable method name."""
        obj = self.ObjectWithGoodMethodName(Mock(), "good_object")
        self.assertTrue(hasattr(obj, "some_method"))

    def test_bad_name_lock(self):
        """Test unacceptable method name."""
        with self.assertRaises(QMI_UsageException):
            self.ObjectWithBadMethodNameLock(Mock(), "bad_object")

    def test_bad_name_unlock(self):
        """Test unacceptable method name."""
        with self.assertRaises(QMI_UsageException):
            self.ObjectWithBadMethodNameUnlock(Mock(), "bad_object")

    def test_bad_name_force_unlock(self):
        """Test unacceptable method name."""
        with self.assertRaises(QMI_UsageException):
            self.ObjectWithBadMethodNameForceUnlock(Mock(), "bad_object")

    def test_bad_name_is_locked(self):
        """Test unacceptable method name."""
        with self.assertRaises(QMI_UsageException):
            self.ObjectWithBadMethodNameIsLocked(Mock(), "bad_object")


class TestRpcObjectLocking(unittest.TestCase):
    def setUp(self):

        # Suppress warnings.
        logging.getLogger("qmi.core.rpc").setLevel(logging.ERROR)

        config = CfgQmi(
            contexts={
                "my_context": CfgContext(tcp_server_port=0)
            }
        )

        self.context = QMI_Context("my_context", config)
        self.context.start()

    def tearDown(self):
        self.context.stop()
        self.context = None

        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_lock(self):
        """Test locking an object."""
        proxy = self.context.make_rpc_object("my_object", MyRpcTestClass)

        result = proxy.lock()
        self.assertTrue(result)
        self.assertTrue(proxy.is_locked())

    def test_lock_with_custom_token(self):
        """Test locking an object with a custom token."""
        proxy = self.context.make_rpc_object("my_object", MyRpcTestClass)

        result = proxy.lock(lock_token="thisismineallmine")
        self.assertTrue(result)
        self.assertTrue(proxy.is_locked())

    def test_lock_unlock(self):
        """Test unlocking an object that was locked within the same proxy."""
        proxy = self.context.make_rpc_object("my_object", MyRpcTestClass)

        result = proxy.lock()
        self.assertTrue(result)
        self.assertTrue(proxy.is_locked())

        result = proxy.unlock()
        self.assertTrue(result)
        self.assertFalse(proxy.is_locked())

    def test_lock_fail(self):
        """Test locking an object from another proxy fails when object is already locked."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")

        result = proxy1.lock()
        self.assertTrue(result)
        self.assertTrue(proxy1.is_locked())
        self.assertTrue(proxy2.is_locked())

        result = proxy2.lock()
        self.assertFalse(result)

    def test_unlock_fail(self):
        """Test unlocking a locked object from other proxy fails if the token is not known."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")

        result = proxy1.lock()
        self.assertTrue(result)
        self.assertTrue(proxy1.is_locked())
        self.assertTrue(proxy2.is_locked())

        result = proxy2.unlock()
        self.assertFalse(result)
        self.assertTrue(proxy1.is_locked())
        self.assertTrue(proxy2.is_locked())

    def test_force_unlock(self):
        """Test force unlocking an object unlocks even if object is locked by another proxy."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")

        result = proxy1.lock()
        self.assertTrue(result)
        self.assertTrue(proxy1.is_locked())
        self.assertTrue(proxy2.is_locked())

        proxy2.force_unlock()
        self.assertFalse(proxy1.is_locked())
        self.assertFalse(proxy2.is_locked())

    def test_lock_unlock_with_custom_token(self):
        """Test unlocking an object with a custom token."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")
        custom_token = "thisismineallmine"

        result = proxy1.lock(lock_token=custom_token)
        self.assertTrue(result)
        self.assertTrue(proxy1.is_locked())
        # Test that unlocking with some other custom token doesn't work
        result = proxy2.unlock(lock_token="precious")
        self.assertFalse(result)
        self.assertTrue(proxy1.is_locked())
        # Then unlock with the custom token
        result = proxy2.unlock(lock_token=custom_token)
        self.assertTrue(result)
        self.assertFalse(proxy2.is_locked())

    def test_lock_with_timeout(self):
        """Test that a proxy object that is locked, can be set pending by another proxy to be locked with a
        timeout. The second proxy achieves the lock when the first one releases it."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")

        result = proxy1.lock()
        self.assertTrue(result)
        # Call release after 0.5 sec
        unlock = Timer(0.5, proxy1.unlock)
        start = time.time()
        unlock.start()

        result = proxy2.lock(timeout=1.0)
        self.assertTrue(result)
        end = time.time()
        self.assertLess(end-start, 0.7)
        # As the context is locked by 2nd context, it should appear as locked for both
        self.assertTrue(proxy1.is_locked())
        self.assertTrue(proxy2.is_locked())

    def test_lock_with_timeout_2(self):
        """The same test as previous, but with the 'client' side now doing the locking."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")

        result = proxy2.lock()
        self.assertTrue(result)
        # Call release after 0.5 sec
        unlock = Timer(0.5, proxy2.unlock)
        start = time.time()
        unlock.start()

        result = proxy1.lock(timeout=1.0)
        self.assertTrue(result)
        end = time.time()
        self.assertLess(end-start, 0.7)
        # As the context is locked by 1st context, it should appear as locked for both
        self.assertTrue(proxy2.is_locked())
        self.assertTrue(proxy1.is_locked())

    def test_lock_with_timeout_fail(self):
        """Test that a proxy object that is locked, can be set pending by another proxy to be locked with a
        timeout. The second proxy time-outs before receiving the proxy."""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy2 = self.context.get_rpc_object_by_name("my_context.my_object")

        result = proxy1.lock()
        self.assertTrue(result)
        # Call release after 0.5 sec
        unlock = Timer(0.5, proxy1.unlock)
        start = time.time()
        unlock.start()

        result = proxy2.lock(timeout=0.4)
        self.assertFalse(result)
        end = time.time()
        self.assertLess(end - start, 0.5)
        time.sleep(0.6 - end + start)
        # The last call was to unlock, so neither should appear locked
        self.assertFalse(proxy1.is_locked())
        self.assertFalse(proxy2.is_locked())

    def test_my_lock_method(self):
        """Test shows that user is allowed to use the keyword argument `lock_token`"""
        proxy1 = self.context.make_rpc_object("my_object", MyRpcTestClass)
        proxy1.my_lock_method(lock_token='foo')

        self.assertTrue(proxy1.lock())
        self.assertTrue(proxy1.is_locked())

        proxy1.my_lock_method(lock_token='foo')

    def test_make_rpc_object_excepts(self):
        """Give an invalid name for a rpc object."""
        with self.assertRaises(QMI_UsageException):
            self.context.make_rpc_object("!@gr$%:(", MyRpcTestClass)


class TestRPCWithClosingService(unittest.TestCase):

    def test_making_object_in_inactive_context_excepts(self):
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0)
            }
        )
        c1 = QMI_Context("c1", config)
        c1.start()
        c1.make_rpc_object("duplicate", MyRpcTestClass)
        with self.assertRaises(QMI_DuplicateNameException):
            c1.make_rpc_object("duplicate", MyRpcTestClass)

    def test_making_duplicate_object_in_context_excepts(self):
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0)
            }
        )
        c1 = QMI_Context("c1", config)
        c1.start()
        c1.stop()
        with self.assertRaises(QMI_InvalidOperationException):
            c1.make_rpc_object("notactive", MyRpcTestClass)

    def test_lock_token_after_closed_context(self):
        """Test locking the object with a lock token and using it from another context with the same name."""
        config = CfgQmi(
            contexts={
                "c1": CfgContext(tcp_server_port=0)
            }
        )

        c1 = QMI_Context("c1", config)
        c1.start()
        c1_port = c1.get_tcp_server_port()
        c1_address = "localhost:{}".format(c1_port)
        # Instantiate class in context c1 as "object provider".
        proxy1 = c1.make_rpc_object("tc1", MyRpcTestClass)
        self.assertFalse(proxy1.is_locked())

        # Now make another context with the same name, and get a proxy
        c2 = QMI_Context("c2", config)
        c2.start()
        c2.connect_to_peer("c1", c1_address)
        # Instantiate class in context 'c1'.
        proxy2 = c2.get_rpc_object_by_name("c1.tc1")
        self.assertFalse(proxy2.is_locked())
        result = proxy2.lock(lock_token="block")
        self.assertTrue(result)
        self.assertTrue(proxy2.is_locked())
        # And then close it
        c2.stop()

        # Then start third context with the same name, obtain the object and unlock it.
        c3 = QMI_Context("c2", config)
        c3.start()
        c3.connect_to_peer("c1", c1_address)
        # Instantiate class in context 'c1'.
        proxy3 = c3.get_rpc_object_by_name("c1.tc1")
        self.assertTrue(proxy3.is_locked())
        result = proxy3.unlock(lock_token="block")
        self.assertTrue(result)
        self.assertFalse(proxy3.is_locked())
        # And then close it
        c3.stop()

        # Then finally stop the "object provider"
        c1.stop()


if __name__ == "__main__":
    unittest.main()



"""
Instrument driver for PtGrey BlackFly cameras via the Aravis library.

This driver depends on Aravis, a free library for video acquisition
from Genicam cameras. Aravis is a C library based on GLib/GObject.

There is no Python module for Aravis.
Instead, Aravis is accessed via the GObject Introspection framework.
Therefore this driver depends on the Python module 'gi'.

Aravis works under Linux and is NOT supported on Windows.

This driver requires that Aravis is installed on the system. When installed
in a non-standard location, the directory containing libaravis-0.6.so must
be added to LD_LIBRARY_PATH.

Aravis must be compiled with introspection enabled, and the Aravis typelib
must be installed on the system. When installed in a non-standard location,
the directory containing Aravis.typelib must be added to GI_TYPELIB_PATH.
"""

import logging
import socket
import struct
import threading
import time
import typing

from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple

import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_UsageException, QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method

# Lazy import of the Aravis module. See the function _import_modules() below.
if typing.TYPE_CHECKING:
    from gi.repository import Aravis
else:
    Aravis = None


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Global mutex protecting the Aravis library.
_aravis_mutex = threading.Lock()


# Camera device information.
CameraInfo = NamedTuple('CameraInfo', [
    ('vendor_name', str),
    ('model_name', str),
    ('serial_number', str),
    ('version', str),
    ('ip_address', str)
])
CameraInfo.__doc__ = """Camera device information."""

# Image acquired from the camera.
ImageInfo = NamedTuple('ImageInfo', [
    ('width', int),
    ('height', int),
    ('offset_x', int),
    ('offset_y', int),
    ('pixel_format', str),
    ('frame_id', int),
    ('image_id', int),
    ('timestamp', float),
    ('gain', float),
    ('black_level', float),
    ('exposure_time', float),
    ('data', np.ndarray)
])
ImageInfo.__doc__ = """Image acquired from the camera.
    :ivar width: Image width in pixels.
    :ivar height: Image height in pixels.
    :ivar offset_x: X offset within sensor area in pixels.
    :ivar offset_y: Y offset within sensor area in pixels.
    :ivar pixel_format: Pixel format description (GenICam PFNC name).
    :ivar frame_id: Frame sequence number.
    :ivar image_id: Unique image ID.
    :ivar timestamp: Timestamp in seconds, or 0.0 if not present.
    :ivar gain: Analog gain in dB, or 0.0 if not present.
    :ivar black_level: Black level in percent, or 0.0 if not present.
    :ivar exposure_time: Exposure time in microseconds, or 0.0 if not present.
    :ivar data: Pixel data as a 2D Numpy array.
"""


def _import_modules() -> None:
    """Import the Aravis library.

    This import is done in a function, instead of at the top-level,
    to avoid an unnecessary dependency for programs that do not access
    the instrument directly.
    """

    global Aravis

    with _aravis_mutex:
        if Aravis is None:
            # Import Aravis introspection module via GObject Introspection
            import gi
            gi.require_version('Aravis', '0.6')
            from gi.repository import Aravis  # pylint: disable=W0621
            _logger.debug("Aravis library version: %s", Aravis._version)  # type: ignore


class PtGrey_BlackFly_Aravis(QMI_Instrument):
    """Instrument driver for PtGrey BlackFly cameras via the Aravis library.

    Although Aravis is a generic library for Genicam cameras, this driver
    is developed specifically for the PtGrey BFLY-GPE-31S4M-C camera.
    It is likely that the driver will also work with other monochrome Ethernet
    Genicam cameras. However modifications will be needed to support USB
    cameras and/or color cameras.

    The driver currently only supports basic image control settings and
    simple image acquisition.
    """

    @staticmethod
    def list_instruments() -> List[CameraInfo]:
        """Return a list of connected camera devices."""

        devices = []

        # Import Aravis.
        _import_modules()

        with _aravis_mutex:

            # Update list of cameras.
            Aravis.update_device_list()

            # Get list of cameras.
            n_devices = Aravis.get_n_devices()
            for i in range(n_devices):
                vendor_name    = Aravis.get_device_vendor(i)
                model_name     = Aravis.get_device_model(i)
                serial_number  = Aravis.get_device_serial_nbr(i)
                ip_address_str = Aravis.get_device_address(i)
                info = CameraInfo(vendor_name=vendor_name,
                                  model_name=model_name,
                                  serial_number=serial_number,
                                  version="",
                                  ip_address=ip_address_str)
                devices.append(info)

        return devices

    def __init__(self, context: QMI_Context, name: str, serial_number: Optional[str] = None, ip_address: Optional[str] = None) -> None:
        """Initialize a driver instance for the specified camera device.

        Either serial_number or ip_address should be specified.

        :argument serial_number: Serial number of camera device.
        :argument ip_address: IP address of camera device.
        """

        super().__init__(context, name)

        if (serial_number is None) and (ip_address is None):
            raise QMI_UsageException("Need either serial_number or ip_address")

        # Import Aravis module.
        _import_modules()

        # Parse IP address.
        if ip_address is None:
            ip_address_int = None
        else:
            (ip_address_int,) = struct.unpack('>I', socket.inet_aton(ip_address))

        # Store camera selection attributes.
        self._serial_number = serial_number
        self._ip_address = ip_address
        self._ip_address_int = ip_address_int

        # Camera object not yet available.
        self._cam = None           # type: Optional[Aravis.Camera]
        self._stream = None        # type: Optional[Aravis.Stream]
        self._chunk_parser = None  # type: Optional[Aravis.ChunkParser]

        # Image callback functions.
        self._image_callbacks = []  # type: List[Callable[[ImageInfo], None]]

    def _convert_image(self, buf: 'Aravis.Buffer') -> ImageInfo:
        """Convert Aravis Buffer object to an ImageInfo instance."""

        if buf.get_status() != Aravis.BufferStatus.SUCCESS:
            raise QMI_InstrumentException("Received incomplete image data")

        # Payload in CHUNK_DATA format does not contain width/height/format
        # meta-data. We can not process such images.
        if buf.get_payload_type() == Aravis.BufferPayloadType.CHUNK_DATA:
            raise QMI_InstrumentException("Received unsupported CHUNK_DATA payload")

        width = buf.get_image_width()
        height = buf.get_image_height()
        offset_x = buf.get_image_x()
        offset_y = buf.get_image_y()
        pixel_format = buf.get_image_pixel_format()
        frame_id = buf.get_frame_id()
        timestamp = buf.get_timestamp()

        # Convert to Numpy array.
        data = buf.get_data()
        if pixel_format == Aravis.PIXEL_FORMAT_MONO_8:
            pixels = np.frombuffer(data, dtype=np.uint8, count=height * width)
        elif pixel_format == Aravis.PIXEL_FORMAT_MONO_16:
            pixels = np.frombuffer(data, dtype='<H', count=height * width)
        elif pixel_format == Aravis.PIXEL_FORMAT_MONO_12_PACKED:
            tmp = np.frombuffer(data, dtype=np.uint8, count=(height * width * 3 + 1) // 2)
            tmp = tmp.astype(np.uint16)
            pixels = np.zeros(height * width, dtype=np.uint16)
            pixels[::2] = ((tmp[1::3] & 0x0f) << 8) | tmp[0::3]
            pixels[1::2] = (tmp[2::3] << 4) | ((tmp[1::3] & 0xf0) >> 4)
        else:
            raise QMI_InstrumentException("Unknown pixel format in image")

        # Reshape to 2D array.
        pixels = pixels.reshape((height, width))

        # Parse chunk data.
        assert self._chunk_parser is not None
        if buf.get_payload_type() in (Aravis.BufferPayloadType.CHUNK_DATA,
                                      Aravis.BufferPayloadType.EXTENDED_CHUNK_DATA):
            gain = self._chunk_parser.get_float_value(buf, "ChunkGain")
            black_level = self._chunk_parser.get_float_value(buf, "ChunkBlackLevel")
            exposure_time = self._chunk_parser.get_float_value(buf, "ChunkExposureTime")
        else:
            gain = 0
            black_level = 0
            exposure_time = 0

        # Convert timestamp to seconds.
        timestamp = 1.0e-9 * timestamp

        return ImageInfo(width=width,
                         height=height,
                         offset_x=offset_x,
                         offset_y=offset_y,
                         pixel_format=pixel_format,
                         frame_id=frame_id,
                         image_id=frame_id,
                         timestamp=timestamp,
                         gain=gain,
                         black_level=black_level,
                         exposure_time=exposure_time,
                         data=pixels)

    def _new_image_cb(self, stream: 'Aravis.Stream') -> None:
        """This function gets called by the Aravis library in a background
        thread when a new image is captured.
        """

        buf = stream.try_pop_buffer()
        if buf is None:
            # Image disappeared before we could handle it.
            return

        if buf.get_status() != Aravis.BufferStatus.SUCCESS:
            _logger.warning("Received incomplete image from camera %s", self._name)
            return

        # Convert image.
        img = self._convert_image(buf)

        # Re-insert buffer into stream.
        stream.push_buffer(buf)

        # Invoke custom callbacks.
        with _aravis_mutex:
            callbacks = self._image_callbacks

        for cb in callbacks:
            try:
                cb(img)
            except:
                _logger.exception("Do not let exception go up to Aravis background thread")

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s serial_number=%s ip_address=%s",
                     self._name, self._serial_number, self._ip_address)
        self._check_is_closed()

        with _aravis_mutex:

            # Update list of cameras.
            Aravis.update_device_list()

            # Scan list of cameras to find our device.
            cam_selected = []
            n_devices = Aravis.get_n_devices()
            for i in range(n_devices):
                device_id         = Aravis.get_device_id(i)
                serial_number     = Aravis.get_device_serial_nbr(i)
                ip_address_str    = Aravis.get_device_address(i)
                (ip_address_int,) = struct.unpack('>I', socket.inet_aton(ip_address_str))
                if (((self._serial_number is None) or (self._serial_number == serial_number))
                        and ((self._ip_address_int is None) or (self._ip_address_int == ip_address_int))):
                    cam_selected.append(device_id)

            if not cam_selected:
                raise QMI_InstrumentException("No camera found matching serial_number={} ip_address={}"
                                              .format(self._serial_number, self._ip_address))
            if len(cam_selected) != 1:
                raise QMI_InstrumentException("Multiple cameras found matching serial_number={} ip_address={}"
                                              .format(self._serial_number, self._ip_address))

            # Open camera device.
            self._cam = Aravis.Camera.new(cam_selected[0])

        # Mark instrument as open.
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        self._check_is_open()
        assert self._cam is not None

        # Stop image acquisition (just in case).
        self._cam.stop_acquisition()

        # Disconnect from camera.
        with _aravis_mutex:
            self._stream = None
            self._chunk_parser = None
            self._cam = None

        # Mark instrument as closed.
        super().close()

    @rpc_method
    def reset(self) -> None:
        """Reset the camera.

        Resetting triggers a complete reboot of the camera and involves
        temporarily closing and re-opening the network connection.
        """
        _logger.info("Resetting %s", self._name)
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        dev.get_feature('DeviceReset').execute()
        self.close()
        time.sleep(10)
        self.open()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return a QMI_InstrumentIdentification instance."""
        self._check_is_open()
        assert self._cam is not None
        vendor_name   = self._cam.get_vendor_name()
        model_name    = self._cam.get_model_name()
        dev = self._cam.get_device()
        serial_number = dev.get_feature('DeviceSerialNumber').get_value_as_string()
        version       = dev.get_feature('DeviceVersion').get_value_as_string()
        return QMI_InstrumentIdentification(vendor=vendor_name,
                                            model=model_name,
                                            serial=serial_number,
                                            version=version)

    @rpc_method
    def get_ip_address(self) -> str:
        """Return IP address of the camera device."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        address = dev.get_device_address()
        ip_address_str = address.get_address().to_string()
        return ip_address_str

    @rpc_method
    def get_device_info(self) -> Dict[str, Any]:
        """Return device information."""
        self._check_is_open()
        assert self._cam is not None

        info = {}

        keys = [
            "DeviceVendorName",
            "DeviceModelName",
            "DeviceVersion",
            "DeviceSerialNumber",
            "DeviceID",
            "DeviceFirmwareVersion",
            "pgrSensorDescription",
            "DeviceTemperature",
            "pgrPowerSupplyVoltage",
            "pgrPowerSupplyCurrent",
        ]

        dev = self._cam.get_device()
        for key in keys:
            f = dev.get_feature(key)
            if f is not None:
                if f.get_value_type().name == "gdouble":
                    v = f.get_value()
                else:
                    v = f.get_value_as_string()
                info[key] = v

        return info

    @rpc_method
    def get_frame_rate(self) -> float:
        """Return the acquisition frame rate in frames per second."""
        self._check_is_open()
        assert self._cam is not None
        return self._cam.get_frame_rate()

    @rpc_method
    def get_frame_rate_auto(self) -> str:
        """Return the automatic frame rate mode ("Off", "Once", or "Continuous")."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        return dev.get_feature('AcquisitionFrameRateAuto').get_string_value()

    @rpc_method
    def set_frame_rate(self, frame_rate: Optional[float] = None, auto_mode: str = "Off") -> None:
        """Set the acquisition frame rate and enable or disable automatic frame rate.

        :argument frame_rate: Frame rate in frame per second.
        :argument auto_mode: Automatic frame rate mode ("Off" or "Continuous").
        """
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        dev.get_feature('AcquisitionFrameRateEnabled').set_value(True)
        dev.get_feature('AcquisitionFrameRateAuto').set_string_value(auto_mode)
        if frame_rate is not None:
            self._cam.set_frame_rate(frame_rate)

    @rpc_method
    def get_gain(self) -> float:
        """Return analog gain in dB."""
        self._check_is_open()
        assert self._cam is not None
        return self._cam.get_gain()

    @rpc_method
    def get_gain_auto(self) -> str:
        """Return the automatic gain mode ("Off", "Once", or "Continuous")."""
        self._check_is_open()
        assert self._cam is not None
        return Aravis.Auto.to_string(self._cam.get_gain_auto())

    @rpc_method
    def set_gain(self, gain: Optional[float] = None, auto_mode: str = "Off") -> None:
        """Set analog gain in dB and enable or disable automatic gain.

        :argument gain: Analog gain in dB.
        :argument auto_mode: Automatic gain mode ("Off", "Once", or "Continuous").
        """
        self._check_is_open()
        assert self._cam is not None
        self._cam.set_gain_auto(Aravis.Auto.from_string(auto_mode))
        if gain is not None:
            self._cam.set_gain(gain)

    @rpc_method
    def get_exposure_mode(self) -> str:
        """Return the exposure mode ("Timed" or "TriggerWidth")."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        return dev.get_feature('ExposureMode').get_string_value()

    @rpc_method
    def set_exposure_mode(self, mode: str) -> None:
        """Set the exposure mode.

        :argument mode: Exposure mode ("Timed" or "TriggerWidth").
        """
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        dev.get_feature('ExposureMode').set_string_value(mode)

    @rpc_method
    def get_exposure_time(self) -> float:
        """Return the exposure time in microseconds."""
        self._check_is_open()
        assert self._cam is not None
        return self._cam.get_exposure_time()

    @rpc_method
    def get_exposure_time_auto(self) -> str:
        """Return the automatic exposure mode ("Off", "Once" or "Continuous")."""
        self._check_is_open()
        assert self._cam is not None
        return Aravis.Auto.to_string(self._cam.get_exposure_time_auto())

    @rpc_method
    def set_exposure_time(self, exposure_time: Optional[float] = None, auto_mode: str = "Off") -> None:
        """Set the exposure time and enable or disable automatic exposure.

        :argument exposure_time: Exposure time in microseconds.
        :argument auto_mode: Automatic exposure mode ("Off", "Once" or "Continuous").
        """
        self._check_is_open()
        assert self._cam is not None
        self._cam.set_exposure_time_auto(Aravis.Auto.from_string(auto_mode))
        if exposure_time is not None:
            self._cam.set_exposure_time(exposure_time)

    @rpc_method
    def get_pixel_format(self) -> str:
        """Return pixel format ("Mono8", "Mono12Packed" or "Mono16")."""
        self._check_is_open()
        assert self._cam is not None
        tbl = {Aravis.PIXEL_FORMAT_MONO_8: "Mono8",
               Aravis.PIXEL_FORMAT_MONO_16: "Mono16",
               Aravis.PIXEL_FORMAT_MONO_12_PACKED: "Mono12Packed"}
        fmt = self._cam.get_pixel_format()
        if fmt not in tbl:
            raise ValueError("Unknown pixel format")
        return tbl[fmt]

    @rpc_method
    def set_pixel_format(self, pixel_format: str) -> None:
        """Set format of the pixel data.

        :argument pixel_format: Format ("Mono8", "Mono12Packed" or "Mono16").
        """
        self._check_is_open()
        assert self._cam is not None
        tbl = {"Mono8": Aravis.PIXEL_FORMAT_MONO_8,
               "Mono16": Aravis.PIXEL_FORMAT_MONO_16,
               "Mono12Packed": Aravis.PIXEL_FORMAT_MONO_12_PACKED}
        if pixel_format not in tbl:
            raise ValueError("Unknown pixel format")
        fmt = tbl[pixel_format]
        self._cam.set_pixel_format(fmt)

    @rpc_method
    def get_image_size(self) -> Tuple[int, int]:
        """Return the current image size.

        :return: Image size as tuple (width, height) in pixels.
        """
        self._check_is_open()
        assert self._cam is not None
        region = self._cam.get_region()
        return (region.width, region.height)

    @rpc_method
    def set_image_size(self, width: int, height: int) -> None:
        """Set the size of images provided by the camera.

        A subset of the sensor area may be selected by setting the image size
        and image offset.

        :argument width: Width of image in pixels (must be a multiple of 4).
        :argument height: Height of image in pixels (must be a multiple of 2).
        """
        self._check_is_open()
        assert self._cam is not None
        region = self._cam.get_region()
        self._cam.set_region(region.x, region.y, width, height)

    @rpc_method
    def get_image_offset(self) -> Tuple[int, int]:
        """Return the current image offset position.

        :return: Offset position as tuple (x, y) in pixels.
        """
        self._check_is_open()
        assert self._cam is not None
        region = self._cam.get_region()
        return (region.x, region.y)

    @rpc_method
    def set_image_offset(self, x: int, y: int) -> None:
        """Set the image offset position.

        :argument x: Horizotal offset from the origin to the area of interest in pixels (must be a multiple of 2).
        :argument y: Vertical offset from the origin to the area of interest in pixels (must be a multiple of 2).
        """
        self._check_is_open()
        assert self._cam is not None
        region = self._cam.get_region()
        self._cam.set_region(x, y, region.width, region.height)

    @rpc_method
    def get_chunk_mode(self) -> Tuple[bool, Dict[str, bool]]:
        """Return the current chunk data configuration.

        :return: A tuple (chunk_data_active, chunk_types_enabled)
            where chunk_data_active is a boolean indicating whether chunk data is active and
            chunk_types_enabled is a dictionary indicating which chunk types are enabled.
        """
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        active = self._cam.get_chunk_mode()
        chunk_types = {}
        for chunk_type in dev.get_feature('ChunkSelector').get_available_string_values():
            dev.get_feature('ChunkSelector').set_string_value(chunk_type)
            chunk_types[chunk_type] = dev.get_feature('ChunkEnable').get_value()
        return (active, chunk_types)

    @rpc_method
    def set_chunk_mode(self, enable: bool) -> None:
        """Enable or disable chunk data mode.

        When chunk data mode is active, the camera adds some meta-data to each
        acquired image.

        In principle, the camera supports separate enabling of specific
        meta-data types. However this function enables all meta-data types
        when chunk data mode is enabled.
        """
        self._check_is_open()
        assert self._cam is not None

        if enable:
            dev = self._cam.get_device()

            # Enable EXTENDED_CHUNK_DATA payload.
            # By default, when chunk mode is enabled, the camera will send
            # CHUNK_DATA payload. Received image buffers in this format do
            # not contain width/height/format meta-data, which are required
            # by our image processing code.
            # By setting the extended_chunk_data_enable bit in the
            # Stream Channel Configuration Register, we ask that the camera
            # send EXTENDED_CHUNK_DATA instead of CHUNK_DATA. This feature
            # is optional and deprecated, but BlackFly cameras support it.
            dev.write_register(0x0d24, 1)

            for chunk_type in dev.get_feature('ChunkSelector').get_available_string_values():
                dev.get_feature('ChunkSelector').set_string_value(chunk_type)
                dev.get_feature('ChunkEnable').set_value(True)

        self._cam.set_chunk_mode(enable)

    @rpc_method
    def get_device_temperature(self) -> float:
        """Return the temperature of the camera in degrees Celcius."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        return dev.get_feature('DeviceTemperature').get_value()

    @rpc_method
    def get_transmit_failure_count(self) -> int:
        """Return the number of failed transmissions since the last reset."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        return dev.get_feature('TransmitFailureCount').get_value()

    @rpc_method
    def get_black_level(self) -> float:
        """Return analog black level in percent."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        return dev.get_feature('BlackLevel').get_value()

    @rpc_method
    def set_black_level(self, level: float) -> None:
        """Set analog black level.

        :argument level: Black level in percent.
        """
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        dev.get_feature('BlackLevel').set_value(level)

    @rpc_method
    def set_indicator_mode(self, enable: bool) -> None:
        """Enable or disable the indicator LED on the camera."""
        self._check_is_open()
        assert self._cam is not None
        dev = self._cam.get_device()
        sval = "Active" if enable else "Inactive"
        dev.get_feature("DeviceIndicatorMode").set_string_value(sval)

    @rpc_method
    def is_streaming(self) -> bool:
        """Return True if the camera is in streaming mode."""
        return self._stream is not None

    @rpc_method
    def start_acquisition(self, num_frames: Optional[int] = None) -> None:
        """Start frame acquisition.

        Set acquisition mode and number of frames as specified, then start
        frame acquisition in the camera.

        Call get_next_image() to fetch each image as it becomes available.
        Call stop_acquisition() to end the acquisition process.

        :argument num_frames: Number of frames to acquire,
            or None to set the camera in Continuous mode.
        """
        self._check_is_open()
        assert self._cam is not None

        _logger.debug("Starting image acquisition on %s num_frames=%r",
                      self._name, num_frames)

        if num_frames is None:
            # Continuous mode.
            self._cam.set_acquisition_mode(Aravis.AcquisitionMode.CONTINUOUS)
        elif num_frames == 1:
            # Single frame.
            self._cam.set_acquisition_mode(Aravis.AcquisitionMode.SINGLE_FRAME)
        elif num_frames > 0 and num_frames < 65536:
            # Multi-frame.
            self._cam.set_acquisition_mode(Aravis.AcquisitionMode.MULTI_FRAME)
            self._cam.set_frame_count(num_frames)
        else:
            raise QMI_UsageException("Invalid value for num_frames")

        # Create stream object and connect to event handler.
        with _aravis_mutex:
            self._chunk_parser = self._cam.create_chunk_parser()
            self._stream = self._cam.create_stream(None, None)
            self._stream.connect("new-buffer", self._new_image_cb)
            if self._image_callbacks:
                self._stream.set_emit_signals(True)
            else:
                self._stream.set_emit_signals(False)

        # Pre-allocate image buffers and give them to the stream to manage.
        payload = self._cam.get_payload()
        num_bufs = min(128, 128*1024*1024 // payload)
        if num_frames is not None:
            num_bufs = min(num_bufs, num_frames)

        for i in range(num_bufs):
            buf = Aravis.Buffer.new(payload)
            self._stream.push_buffer(buf)

        # Start image acquisition.
        self._cam.start_acquisition()

    @rpc_method
    def stop_acquisition(self) -> None:
        """Stop frame acquisition in the camera."""
        self._check_is_open()
        assert self._cam is not None
        _logger.debug("Stopping image acquisition on %s", self._name)

        # Stop image acquisition.
        self._cam.stop_acquisition()

        # Clean up stream object.
        if self._stream is not None:
            self._stream.set_emit_signals(False)
        self._stream = None
        self._chunk_parser = None

    @rpc_method
    def get_next_image(self, timeout: Optional[float] = None) -> ImageInfo:
        """Return the next image from the current acquisition.

        This function blocks until the next image is available.
        It then returns a named tuple containing image attributes
        as well as pixel data (as a Numpy array). The image is
        then released from the PySpin buffer.

        This function must only be used when the camera is in
        acquisition mode (after start_acquisition()).

        This function must only be called if a next image is expected
        (based on the num_frames parameter to start_acquisition()).
        """
        self._check_is_open()
        assert self._cam is not None

        if self._stream is None:
            raise QMI_InstrumentException("Camera not in streaming mode")

        if timeout is None:
            buf = self._stream.pop_buffer()
        else:
            buf = self._stream.timeout_pop_buffer(int(1000000 * timeout))

        if buf is None:
            raise QMI_TimeoutException("Timeout while waiting for image")

        if buf.get_status() != Aravis.BufferStatus.SUCCESS:
            raise QMI_InstrumentException("Received incomplete image data")

        img = self._convert_image(buf)

        self._stream.push_buffer(buf)

        return img

    def register_image_callback(self, callback: Callable[[ImageInfo], None]) -> None:
        """Register callback function to be invoked when a new image becomes available.

        Note that the callback function may be invoked from a background thread.
        """
        with _aravis_mutex:
            self._image_callbacks.append(callback)
            if self._stream is not None:
                self._stream.set_emit_signals(True)

    def unregister_image_callback(self, callback: Callable) -> None:
        """Unregister the specified image callback function."""
        with _aravis_mutex:
            self._image_callbacks.remove(callback)



"""WL Photonics, wavelength-tunable filter.

The qmi.instruments.wl_photonics package provides support for:
  - WLTF-N: Narrowband USB wavelength-tunable filters.
"""
from qmi.instruments.wl_photonics.wltf_n import WlPhotonics_WltfN



"""
Instrument driver for the Tektronix AFG31000 series arbitrary function generator.
"""

import enum
import logging
import math
import re
from typing import Tuple, Type, TypeVar, Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport


# Type variable.
_EnumTypeVar = TypeVar("_EnumTypeVar", bound=enum.Enum)

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Waveform(enum.Enum):
    """Waveform shape."""
    SINE = "SIN"
    SQUARE = "SQU"
    PULSE = "PULS"
    RAMP = "RAMP"
    NOISE = "PRN"
    DC = "DC"
    SINC = "SINC"
    GAUSS = "GAUS"
    LORENTZ = "LOR"
    EXP_RISE = "ERIS"
    EXP_DECAY = "EDEC"
    HAVERSINE = "HAV"
    EMEMORY = "EMEM"
    EFILE = "EFIL"


class BurstMode(enum.Enum):
    """Burst mode."""
    TRIGGERED = "TRIG"
    GATED = "GAT"


class TriggerEdge(enum.Enum):
    """Trigger edge."""
    RISING = "POS"
    FALLING = "NEG"


class Tektronix_AFG31000(QMI_Instrument):
    """Instrument driver for the Tektronix AFG31000 series arbitrary function generator.

    This driver has been tested with the AFG31022 but probably works with all AFG31000 series instruments.

    This driver communicates with the instrument via USB or via Ethernet.
    """

    # Response timeout in seconds.
    RESPONSE_TIMEOUT = 5.0

    # The instrument returns a value above this threshold to indicate infinity.
    _INFINITY_THRESHOLD = 1.0e30

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        To connect to the instrument via Ethernet, specify transport descriptor
          "tcp:<IP-address>".

        To connect to the instrument via USB, specify transport descriptor
          "usbtmc:vendorid=0x0699:productid=0x0356:serialnr=....".

        Parameters:
            context:    QMI context.
            name:       Name for this instrument instance.
            transport:  Transport descriptor to access this instrument.
        """

        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"port": 5025})
        self._scpi_transport = ScpiProtocol(self._transport, default_timeout=self.RESPONSE_TIMEOUT)

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._transport.open()
        try:
            # Clear error queue to avoid reporting stale errors.
            self._scpi_transport.write("*CLS")
        except OSError:
            self._transport.close()
            raise
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _ask_int(self, cmd: str) -> int:
        """Query an integer value from the instrument."""
        resp = self._scpi_transport.ask(cmd)
        resp = resp.strip()
        try:
            return int(resp)
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response to {cmd!r}: {resp!r}")

    def _ask_float(self, cmd: str) -> float:
        """Query a floating point value from the instrument."""
        resp = self._scpi_transport.ask(cmd)
        resp = resp.strip()
        try:
            return float(resp)
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response to {cmd!r}: {resp!r}")

    def _ask_enum(self, cmd: str, typ: Type[_EnumTypeVar]) -> _EnumTypeVar:
        """Query a mnemonic response from the instrument."""
        resp = self._scpi_transport.ask(cmd)
        resp = resp.strip()
        try:
            return typ(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response to {cmd!r}: {resp!r}")

    @staticmethod
    def _check_channel(channel: int) -> None:
        """Check that the channel number is valid."""
        if channel not in (1, 2):
            raise ValueError(f"Invalid channel number {channel!r}")

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""

        # Wait until previous command finished.
        self._scpi_transport.ask("*OPC?")

        # Read error queue.
        resp = self._scpi_transport.ask("SYST:ERR?")
        if not re.match(r"^\s*[+-]?0\s*,", resp):
            # An error occurred. Clear the error queue, in case there are more pending errors.
            self._scpi_transport.write("*CLS")
            # Report the error.
            raise QMI_InstrumentException(f"Instrument returned error: {resp}")

    @rpc_method
    def reset(self) -> None:
        """Reset most instrument settings to factory defaults.

        This function is equivalent to the "default" button on the front panel.
        """
        self._scpi_transport.write("*CLS")
        self._scpi_transport.write("*RST")
        self._check_error()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        resp = self._scpi_transport.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to '*IDN?': {resp!r}")
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def get_external_reference_enabled(self) -> bool:
        """Return True if the external reference clock input is selected."""
        class RefSource(enum.Enum):
            """Enum for type of RF reference source."""
            INTERNAL = "INT"
            EXTERNAL = "EXT"
        refsource = self._ask_enum("SOURCE:ROSCILLATOR:SOURCE?", RefSource)
        return refsource == RefSource.EXTERNAL

    @rpc_method
    def set_external_reference_enabled(self, enable: bool) -> None:
        """Choose internal or external reference clock source.

        Parameters:
            enable: True to select the external 10 MHz reference clock input.
                    False to select the internal reference source.
        """
        arg = "EXT" if enable else "INT"
        self._scpi_transport.write(f"SOURCE:ROSCILLATOR:SOURCE {arg}")
        self._check_error()

    @rpc_method
    def get_output_enabled(self, channel: int) -> bool:
        """Return True if the specified output is enabled.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            True if the specified output is enabled, False if disabled.
        """
        self._check_channel(channel)
        resp = self._ask_int(f"OUTPUT{channel}:STATE?")
        return resp != 0

    @rpc_method
    def set_output_enabled(self, channel: int, enable: bool) -> None:
        """Enable or disable the specified output channel.

        Parameters:
            channel: Output channel (1 or 2).
            enable:  True to enable the output, False to disable.
        """
        self._check_channel(channel)
        arg = 1 if enable else 0
        self._scpi_transport.write(f"OUTPUT{channel}:STATE {arg}")
        self._check_error()

    @rpc_method
    def get_waveform(self, channel: int) -> Waveform:
        """Return the current waveform for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            A `Waveform` constant.
        """
        self._check_channel(channel)
        return self._ask_enum(f"SOURCE{channel}:FUNCTION:SHAPE?", Waveform)

    @rpc_method
    def set_waveform(self, channel: int, waveform: Waveform) -> None:
        """Set the waveform for the specified channel.

        Selecting a waveform that is not compatible with the current modulation type,
        will cause the modulation type to be set to `Continuous`.

        Parameters
            channel: Output channel (1 or 2).
            waveform: A `Waveform` constant specifying the new waveform.
                      Waveforms `EMEMORY` and `EFILE` are currently not supported by this driver.
        """
        self._check_channel(channel)
        arg = Waveform(waveform).value
        self._scpi_transport.write(f"SOURCE{channel}:FUNCTION:SHAPE {arg}")
        self._check_error()

    @rpc_method
    def get_sweep_enabled(self, channel: int) -> bool:
        """Return True if frequency sweeping is enabled for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            True if frequency sweeping is enabled, False if disabled.
        """
        class FrequencyMode(enum.Enum):
            """Enum for type of frequency mode."""
            CW = "CW"
            SWEEP = "SWE"
        self._check_channel(channel)
        mode = self._ask_enum(f"SOURCE{channel}:FREQUENCY:MODE?", FrequencyMode)
        return mode == FrequencyMode.SWEEP

    @rpc_method
    def get_continuous_frequency(self, channel: int) -> float:
        """Return the frequency of the output waveform.

        This command is not available when frequency sweeping is enabled.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            Output frequency in Hz.
        """
        self._check_channel(channel)
        return self._ask_float(f"SOURCE{channel}:FREQUENCY:CW?")

    @rpc_method
    def set_continuous_frequency(self, channel: int, frequency: float) -> None:
        """Set the channel to fixed frequency mode with the specified frequency.

        This command disables frequency sweeping and enables fixed frequency (continuous) mode.

        Parameters:
            channel: Output channel (1 or 2).
            frequency: New output frequency in Hz.
        """
        self._check_channel(channel)
        self._scpi_transport.write(f"SOURCE{channel}:FREQUENCY:MODE CW")
        self._scpi_transport.write(f"SOURCE{channel}:FREQUENCY:CW {frequency:.6f}")
        self._check_error()

    @rpc_method
    def get_amplitude(self, channel: int) -> Tuple[float, float]:
        """Return the current output amplitude levels.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            Tuple (low, high) of output levels in Volt.
        """
        self._check_channel(channel)
        vlow = self._ask_float(f"SOURCE{channel}:VOLTAGE:LEVEL:LOW?")
        vhigh = self._ask_float(f"SOURCE{channel}:VOLTAGE:LEVEL:HIGH?")
        return (vlow, vhigh)

    @rpc_method
    def set_amplitude(self, channel: int, vlow: float, vhigh: float) -> None:
        """Set the output amplitude levels for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).
            vlow:    Low output amplitude in Volt.
            vhigh:   High output amplitude in Volt.
        """
        self._check_channel(channel)
        self._scpi_transport.write(f"SOURCE{channel}:VOLTAGE:LEVEL:LOW {vlow:.4f}")
        self._scpi_transport.write(f"SOURCE{channel}:VOLTAGE:LEVEL:HIGH {vhigh:.4f}")
        self._check_error()

    @rpc_method
    def get_pulse_duty_cycle(self, channel: int) -> float:
        """Return the current pulse duty cycle for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            Duty cycle as a percentage between 0.0 and 100.0.
        """
        self._check_channel(channel)
        return self._ask_float(f"SOURCE{channel}:PULSE:DCYCLE?")

    @rpc_method
    def set_pulse_duty_cycle(self, channel: int, duty_cycle: float) -> None:
        """Set the pulse duty cycle for the specified channel.

        This function also configures the instrument to hold the duty cycle fixed
        when the pulse frequency is changed (thus changing the pulse width).

        Parameters:
            channel: Output channel (1 or 2).
            duty_cycle: New duty cycle as a percentage between 0.0 and 100.0.
                        The valid range also depends on the pulse frequency
                        and the transition times.
        """
        self._check_channel(channel)
        self._scpi_transport.write(f"SOURCE{channel}:PULSE:HOLD DUTY")
        self._scpi_transport.write(f"SOURCE{channel}:PULSE:DCYCLE {duty_cycle:.6f}")
        self._check_error()

    @rpc_method
    def get_pulse_delay(self, channel: int) -> float:
        """Return the current pulse delay for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            Pulse lead delay in seconds.
        """
        self._check_channel(channel)
        return self._ask_float(f"SOURCE{channel}:PULSE:DELAY?")

    @rpc_method
    def set_pulse_delay(self, channel: int, delay: float) -> None:
        """Set pulse delay for the specified channel.

        The minimum allowed pulse delay is 0.
        The maximum allowed pulse delay depends on the pulse period and duty cycle.
        The pulse delay must be less than the pulse period.
        In burst mode, the delayed pulse must end before a non-delayed next pulse would have started.

        Parameters:
            channel: Output channel (1 or 2).
            delay:   New pulse lead delay in seconds.
        """
        self._check_channel(channel)
        self._scpi_transport.write(f"SOURCE{channel}:PULSE:DELAY {delay:.11f}")
        self._check_error()

    @rpc_method
    def get_burst_mode(self, channel: int) -> BurstMode:
        """Return the current burst mode for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            `BurstMode.TRIGGERED` or `BurstMode.GATED`.
        """
        self._check_channel(channel)
        return self._ask_enum(f"SOURCE{channel}:BURST:MODE?", BurstMode)

    @rpc_method
    def set_burst_mode(self, channel: int, mode: BurstMode) -> None:
        """Set the burst mode of the specified channel.

        Parameters:
            channel: Output channel (1 or 2).
            mode:    Burst mode (`BurstMode.TRIGGERED` or `BurstMode.GATED`).
        """
        self._check_channel(channel)
        arg = BurstMode(mode).value
        self._scpi_transport.write(f"SOURCE{channel}:BURST:MODE {arg}")
        self._check_error()

    @rpc_method
    def get_burst_count(self, channel: int) -> Union[int, float]:
        """Return the number of cycles in burst mode.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            Number of cycles per output burst, or `math.inf` for infinite cycles.
        """
        self._check_channel(channel)
        val = self._ask_float(f"SOURCE{channel}:BURST:NCYCLES?")
        if val > self._INFINITY_THRESHOLD:
            return math.inf
        return int(val)

    @rpc_method
    def set_burst_count(self, channel: int, cycles: Union[int, float]) -> None:
        """Set the number of cycles in burst mode.

        Parameters:
            channel: Output channel (1 or 2).
            cycles:  Number of cycles, range 1 to 10**6, or `math.inf` for infinite cycles.
        """
        self._check_channel(channel)
        arg = "INF" if math.isinf(cycles) else str(cycles)
        self._scpi_transport.write(f"SOURCE{channel}:BURST:NCYCLES {arg}")
        self._check_error()

    @rpc_method
    def get_burst_enabled(self, channel: int) -> bool:
        """Return True if burst mode is enabled for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            True if burst mode is enabled for the specified channel, False if disabled.
        """
        self._check_channel(channel)
        resp = self._ask_int(f"SOURCE{channel}:BURST:STATE?")
        return resp != 0

    @rpc_method
    def set_burst_enabled(self, channel: int, enable: bool) -> None:
        """Enable or disable burst mode for the specified channel.

        Parameters:
            channel: Output channel (1 or 2).
            enable:  True to enable burst mode, False to disable.
        """
        self._check_channel(channel)
        arg = 1 if enable else 0
        self._scpi_transport.write(f"SOURCE{channel}:BURST:STATE {arg}")
        self._check_error()

    @rpc_method
    def get_external_trigger_enabled(self) -> bool:
        """Return True if the external trigger input is enabled."""
        class TriggerSource(enum.Enum):
            """Enum for type of trigger source."""
            TIMER = "TIM"
            EXTERNAL = "EXT"
        trigger = self._ask_enum("TRIGGER:SOURCE?", TriggerSource)
        return trigger == TriggerSource.EXTERNAL

    @rpc_method
    def set_external_trigger_enabled(self, enable: bool) -> None:
        """Enable or disable the external trigger input.

        Parameters:
            enable: True to select the external trigger input.
                    False to select the internal timer as trigger source.
        """
        arg = "EXT" if enable else "TIM"
        self._scpi_transport.write(f"TRIGGER:SOURCE {arg}")
        self._check_error()

    @rpc_method
    def get_trigger_edge(self) -> TriggerEdge:
        """Return the current trigger edge.

        Returns:
            `TriggerEdge.RISING` or `TriggerEdge.FALLING`.
        """
        return self._ask_enum("TRIGGER:SLOPE?", TriggerEdge)

    @rpc_method
    def set_trigger_edge(self, edge: TriggerEdge) -> None:
        """Set the trigger edge.

        Parameters:
            edge: Either `TriggerEdge.RISING` or `TriggerEdge.FALLING`.
        """
        arg = TriggerEdge(edge).value
        self._scpi_transport.write(f"TRIGGER:SLOPE {arg}")
        self._check_error()

    @rpc_method
    def get_output_load_impedance(self, channel: int) -> float:
        """Return the configured load impedance of the specified output channel.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            Assumed load impedance in Ohm, or `math.inf` for open circuit.
        """
        self._check_channel(channel)
        imp = self._ask_float(f"OUTPUT{channel}:IMPEDANCE?")
        if imp > self._INFINITY_THRESHOLD:
            return math.inf
        return imp

    @rpc_method
    def set_output_load_impedance(self, channel: int, impedance: float) -> None:
        """Configure the load impedance of the specified output channel.

        Note: The `output impedance` of the instrument is fixed at 50 Ohm.
        This command configures the impedance of the load that is assumed to
        be connected to the output. This affects the calculation of output amplitudes.

        Parameters:
            channel: Output channel (1 or 2).
            impedance: Load impedance in Ohm, range 1 to 10000, or `math.inf` to assume open circuit.
        """
        self._check_channel(channel)
        arg = "INF" if (impedance > self._INFINITY_THRESHOLD) else f"{impedance:.1f}"
        self._scpi_transport.write(f"OUTPUT{channel}:IMPEDANCE {arg}")
        self._check_error()

    @rpc_method
    def get_output_inverted(self, channel: int) -> bool:
        """Return True if the output polarity of the specified channel is inverted.

        Parameters:
            channel: Output channel (1 or 2).

        Returns:
            True if polarity is inverted, False if polarity is normal.
        """
        class Polarity(enum.Enum):
            """Enum for polarity of output."""
            NORMAL = "NORM"
            INVERTED = "INV"
        self._check_channel(channel)
        polarity = self._ask_enum(f"OUTPUT{channel}:POLARITY?", Polarity)
        return polarity == Polarity.INVERTED

    @rpc_method
    def set_output_inverted(self, channel: int, enable: bool) -> None:
        """Enable or disable inverted output polarity.

        Parameters:
            channel: Output channel (1 or 2).
            enable:  True to select inverted output, False to select normal output polarity.
        """
        self._check_channel(channel)
        arg = "INV" if enable else "NORM"
        self._scpi_transport.write(f"OUTPUT{channel}:POLARITY {arg}")
        self._check_error()

    @rpc_method
    def set_display_brightness(self, brightness: float) -> None:
        """Set brightness of the touchscreen display.

        Parameters:
            brightness: Display brightness in range 0.0 to 1.0.
                        Value 0.0 is the lowest brightness but does not
                        completely disable the display.
        """
        self._scpi_transport.write(f"DISPLAY:BRIGHTNESS {brightness:.3f}")
        self._check_error()
